# Copyright 2019 Toyota Research Institute. All rights reserved.
"""
Script for generating protocol files from
input parameters and procedure templates

Usage:
    generate_protocol [INPUT_JSON]

Options:
    -h --help        Show this screen
    --version        Show version


The `generate_protocol` script will generate a protocol file from input
parameters defined in the rows of a CSV-formatted input file.
It stores its outputs in `/data-share/protocols/`|
                                                |-`/procedures/`
                                                |-`/schedules/`
                                                |-`/names/`
For Maccor procedures the output procedures will be stored in `/data-share/protocols/procedures/`
For Arbin schedules the schedules will be stored in `/data-share/protocols/schedules/`
Additionally a file containing the names for the test files generated during the function call
will be stored in `/data-share/protocols/names/` with a date and time in the name to differentiate it
This file is to facilitate the process of starting tests on the cycling machines, by making it easier to
enter data into the appropriate fields.

The input json must contain the following fields
* `file_list` - filenames/paths corresponding to the input csv files

The output json will contain the following fields
* `file_list` - list of protocol files corresponding to the input files

Example:
$ generate_protocol '{"file_list": ["/data-share/raw/parameters/procedure_params_000112e3.csv"]}'
{
    "file_list": ["/data-share/protocols/procedures/name_1.000", "/data-share/protocols/procedures/name_2.000"]
}
"""

import os
import warnings
import json
import datetime
import csv
import numpy as np
import pandas as pd
from docopt import docopt
from monty.serialization import loadfn

from beep import logger, __version__
from beep.protocol import PROCEDURE_TEMPLATE_DIR
from beep.protocol.maccor import Procedure


from beep.utils import KinesisEvents
s = {'service': 'ProtocolGenerator'}


def convert_velocity_to_power_waveform(waveform_file, velocity_units):
    """
    Helper function to perform model based conversion of velocity waveform into power waveform.

    For model description and parameters ref JECS, 161 (14) A2099-A2108 (2014)
    "Model-Based SEI Layer Growth and Capacity Fade Analysis for EV and PHEV Batteries and Drive Cycles"

    Args:
        waveform_file (str): file containing tab or comma delimited values of time and velocity
        velocity_units (str): units of velocity. Accept 'mph' or 'kmph' or 'mps'

    returns
    pd.DataFrame with two columns: time (sec) and power (W). Negative = Discharge
    """
    df = pd.read_csv(waveform_file, sep='\t', header=0)
    df.columns = ['t', 'v']

    if velocity_units == 'mph':
        scale = 1600.0 / 3600.0
    elif velocity_units == 'kmph':
        scale = 1000.0 / 3600.0
    elif velocity_units == 'mps':
        scale = 1.0
    else:
        raise NotImplementedError

    df.v = df.v * scale

    # Define model constants
    m = 1500  # kg
    rolling_resistance_coef = 0.01  # rolling resistance coeff
    g = 9.8  # m/s^2
    theta = 0  # gradient in radians
    rho = 1.225  # kg/m^3
    drag_coef = 0.34  # Coeff of drag
    frontal_area = 1.75  # m^2
    v_wind = 0  # wind velocity in m/s

    # Power = Force * vel
    # Force = Rate of change of momentum + Rolling frictional force + Aerodynamic drag force

    # Method treats the time-series as is and does not interpolate on a uniform grid before computing gradient.
    power = m * np.gradient(df.v, df.t) + rolling_resistance_coef * m * g * np.cos(
        theta * np.pi / 180) + 0.5 * rho * drag_coef * frontal_area * (df.v - v_wind) ** 2

    power = -power * df.v  # positive power = charge

    return pd.DataFrame({'time': df.t,
                         'power': power})


def generate_protocol_files_from_csv(csv_filename, output_directory=None):

    """
    Generates a set of protocol files from csv filename input by
    reading protocol file input corresponding to each line of
    the csv file. Writes a csv file that.

    Args:
        csv_filename (str): CSV containing protocol file parameters.
        output_directory (str): directory in which to place the output files
    """
    # Read csv file
    protocol_params_df = pd.read_csv(csv_filename)

    new_files = []
    names = []
    result = ''
    message = {'comment': '',
               'error': ''}
    if output_directory is None:
        output_directory = PROCEDURE_TEMPLATE_DIR
    for index, protocol_params in protocol_params_df.iterrows():
        template = protocol_params['template']

        # Switch for template invocation
        if template == "EXP.000":
            procedure = Procedure.from_exp(
                **protocol_params[["cutoff_voltage", "charge_rate", "discharge_rate"]]
            )
        elif template == 'diagnosticV2.000':
            diag_params_df = pd.read_csv(os.path.join(PROCEDURE_TEMPLATE_DIR,
                                                      "PreDiag_parameters - DP.csv"))
            diagnostic_params = diag_params_df[diag_params_df['diagnostic_parameter_set'] ==
                                               protocol_params['diagnostic_parameter_set']].squeeze()

            # TODO: should these be separated?
            procedure = Procedure.from_regcyclev2(
                protocol_params
            )
            procedure.add_procedure_diagcyclev2(
                protocol_params["capacity_nominal"], diagnostic_params
            )

        # TODO: how are these different?
        elif template in ['diagnosticV3.000', 'diagnosticV4.000']:
            diag_params_df = pd.read_csv(os.path.join(PROCEDURE_TEMPLATE_DIR,
                                                      "PreDiag_parameters - DP.csv"))
            diagnostic_params = diag_params_df[diag_params_df['diagnostic_parameter_set'] ==
                                               protocol_params['diagnostic_parameter_set']].squeeze()

            procedure = Procedure.generate_procedure_regcyclev3(index, protocol_params)
            procedure.generate_procedure_diagcyclev3(
                    protocol_params["capacity_nominal"], diagnostic_params
            )
        else:
            warnings.warn("Unsupported file template {}, skipping.".format(template))
            result = "error"
            message = {'comment': 'Unable to find template: ' + template,
                       'error': 'Not Found'}
            continue

        filename_prefix = '_'.join(
            [protocol_params["project_name"], '{:06d}'.format(protocol_params["seq_num"])])
        filename = "{}.000".format(filename_prefix)
        filename = os.path.join(output_directory, 'procedures', filename)
        logger.info(filename, extra=s)
        if not os.path.isfile(filename):
            procedure.to_file(filename)
            new_files.append(filename)
            names.append(filename_prefix + '_')

        elif '.sdu' in template:
            logger.warning('Schedule file generation not yet implemented', extra=s)
            result = "error"
            message = {'comment': 'Schedule file generation is not yet implemented',
                       'error': 'Not Implemented'}

    # This block of code produces the file containing all of the run file
    # names produced in this function call. This is to make starting tests easier
    _, namefile = os.path.split(csv_filename)
    namefile = namefile.split('_')[0] + '_names_'
    namefile = namefile + datetime.datetime.now().strftime("%Y%m%d_%H%M") + '.csv'
    with open(os.path.join(output_directory, "names", namefile), 'w', newline='') as outputfile:
        wr = csv.writer(outputfile)
        for name in names:
            wr.writerow([name])
    outputfile.close()

    if not result:
        result = "success"
        message = {'comment': 'Generated {} protocols'.format(str(len(new_files))),
                   'error': ''}

    return new_files, result, message


def process_csv_file_list_from_json(
        file_list_json,
        processed_dir='data-share/protocols/'):
    """

    Args:
        file_list_json (str):
        processed_dir (str):

    Returns:
        str:
    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup Events
    events = KinesisEvents(service='ProtocolGenerator', mode=file_list_data['mode'])

    file_list = file_list_data['file_list']
    all_output_files = []
    protocol_dir = os.path.join(os.environ.get("BEEP_PROCESSING_DIR", "/"),
                              processed_dir)
    for filename in file_list:
        output_files, result, message = generate_protocol_files_from_csv(
            filename, output_directory=protocol_dir)
        all_output_files.extend(output_files)

    output_data = {"file_list": all_output_files,
                   "result": result,
                   "message": message
                   }

    events.put_generate_event(output_data, "complete")

    return json.dumps(output_data)


def main():
    """Main function for the script"""
    logger.info('starting', extra=s)
    logger.info('Running version=%s', __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args['INPUT_JSON']
        print(process_csv_file_list_from_json(input_json), end="")
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info('finish', extra=s)
    return None


if __name__ == "__main__":
    main()



"""
Configuration file for various runtime environments for BEEP-EP.


"""

config = {

    'local': {
        'logging': {
            'container': 'Testing',
            'streams': ['file']
        },
        'kinesis': {
            'stream': 'local/beep/eventstream'
        }
    },

    'dev': {
        'logging': {
            'container': 'Testing',
            'streams': ['file']
        },
        'kinesis': {
            'stream': 'local/beep/eventstream'
        }
    },

    'test': {
        'logging': {
            'container': 'Testing',
            'streams': ['file']
        },
        'kinesis': {
            'stream': 'local/beep/eventstream'
        }
    },

    'stage': {
        'logging': {
            'container': 'BEEP_EP',
            'streams': ['CloudWatch', 'stdout']
        },
        'kinesis': {
            'stream': 'stage/beep/eventstream/stage'
        }
    },

    'prod': {}

}



# Copyright 2019 Toyota Research Institute. All rights reserved.
"""
Module and scripts for generating descriptors (quantities listed
in cell_analysis.m) from cycle-level summary statistics.

Usage:
    featurize [INPUT_JSON]

Options:
    -h --help        Show this screen
    --version        Show version


The `featurize` script will generate features according to the methods
contained in beep.featurize.  It places output files corresponding to
features in `/data-share/features/`.

The input json must contain the following fields

* `file_list` - a list of processed cycler runs for which to generate features

The output json file will contain the following:

* `file_list` - a list of filenames corresponding to the locations of the features

Example:
```angular2
$ featurize '{"invalid_file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
    "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv"],
    "file_list": ["/data-share/structure/FastCharge_2_CH29_structure.json"]}'
{"file_list": ["/data-share/features/FastCharge_2_CH29_full_model_features.json"]}
```
"""

import os
import json
import numpy as np
import pandas as pd
import math
from abc import ABCMeta, abstractmethod
from docopt import docopt
from monty.json import MSONable
from monty.serialization import loadfn, dumpfn
from scipy.stats import skew, kurtosis
from beep.collate import scrub_underscore_suffix, add_suffix_to_filename
from beep.utils import KinesisEvents
from beep.helpers import featurizer_helpers
from beep import logger, ENVIRONMENT, __version__
from beep.structure import get_protocol_parameters

s = {'service': 'DataAnalyzer'}


class BeepFeatures(MSONable, metaclass=ABCMeta):
    """
    Class corresponding to feature baseline feature object.
    Attributes:
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    class_feature_name = 'Base'

    def __init__(self, name, X, metadata):
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def from_run(cls, input_filename, feature_dir, processed_cycler_run):
        """
        This method contains the workflow for the creation of the feature class
        Since the workflow should be the same for all of the feature classed this
        method should not be overridden in any of the derived classes. If the class
        can be created (feature generation succeeds, etc.) then the class is returned.
        Otherwise the return value is False
        Args:
            input_filename (str): path to the input data from processed cycler run
            feature_dir (str): path to the base directory for the feature sets.
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            (beep.featurize.BeepFeatures): class object for the feature set
        """
        if cls.validate_data(processed_cycler_run):
            output_filename = cls.get_feature_object_name_and_path(input_filename, feature_dir)
            feature_object = cls.features_from_processed_cycler_run(processed_cycler_run)
            metadata = cls.metadata_from_processed_cycler_run(processed_cycler_run)
            return cls(output_filename, feature_object, metadata)
        else:
            return False

    @classmethod
    @abstractmethod
    def validate_data(cls, processed_cycler_run):
        raise NotImplementedError

    @classmethod
    def get_feature_object_name_and_path(cls, input_path, feature_dir):
        """
        This function determines how to name the object for a specific feature class
        and creates the full path to save the object. This full path is also used as
        the feature name attribute
        Args:
            input_path (str): path to the input data from processed cycler run
            feature_dir (str): path to the base directory for the feature sets.
        Returns:
            str: the full path (including filename) to use for saving the feature
                object
        """
        new_filename = os.path.basename(input_path)
        new_filename = scrub_underscore_suffix(new_filename)

        # Append model_name along with "features" to demarcate
        # different models when saving the feature vectors.
        new_filename = add_suffix_to_filename(new_filename,
                                              "_features" + "_" + cls.class_feature_name)
        if not os.path.isdir(os.path.join(feature_dir, cls.class_feature_name)):
            os.makedirs(os.path.join(feature_dir, cls.class_feature_name))
        feature_path = os.path.join(feature_dir, cls.class_feature_name, new_filename)
        feature_path = os.path.abspath(feature_path)
        return feature_path

    @classmethod
    @abstractmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run):
        raise NotImplementedError

    @classmethod
    def metadata_from_processed_cycler_run(cls, processed_cycler_run):
        metadata = {
            'barcode': processed_cycler_run.barcode,
            'protocol': processed_cycler_run.protocol,
            'channel_id': processed_cycler_run.channel_id
        }
        return metadata

    def as_dict(self):
        """
        Method for dictionary serialization
        Returns:
            dict: corresponding to dictionary for serialization
        """
        obj = {"@module": self.__class__.__module__,
               "@class": self.__class__.__name__,
               "name": self.name,
               "X": self.X.to_dict("list"),
               "metadata": self.metadata
               }
        return obj

    @classmethod
    def from_dict(cls, d):
        """MSONable deserialization method"""
        d['X'] = pd.DataFrame(d['X'])
        return cls(**d)


class DiagnosticCyclesFeatures(BeepFeatures):
    """
    Object corresponding to feature object. Includes constructors
    to create the features, object names and metadata attributes in the
    object
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    # Class name for the feature object
    class_feature_name = 'DiagnosticCyclesFeatures'
    diagnostic_cycle_types = ['reset', 'hppc', 'rpt_0.2C', 'rpt_1C', 'rpt_2C']

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            feature_object (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        conditions = []
        if not hasattr(processed_cycler_run, 'diagnostic_summary') & hasattr(processed_cycler_run, 'diagnostic_interpolated'):
            return False
        if processed_cycler_run.diagnostic_summary.empty:
            return False
        else:
            conditions.append(set(cls.diagnostic_cycle_types) ==
                              set(processed_cycler_run.diagnostic_summary.cycle_type.unique()))
            conditions.append(cls.check_relaxation_features_viable(processed_cycler_run))

        return all(conditions)

    @classmethod
    def check_relaxation_features_viable(cls, processed_cycler_run, n_soc_windows=8):
        """
        This function returns if it is viable to compute the relaxation features. Will return True if
        all the SOC windows for the HPPC are there for both the 1st and 2nd diagnostic cycles, and False
        if otherwise.

        Args:
            processed_cycler_run(beep.structure.ProcessedCyclerRun)
            n_soc_windows (int): threshold number of soc windows a cell must have


        Returns:
            (boolean): True if all SOC window available in both diagnostic cycles. False otherwise.
        """
        conditions_met = []
        # chooses the first and the second diagnostic cycle
        for hppc_chosen in [0, 1]:

            # Getting just the HPPC cycles
            hppc_diag_cycles = processed_cycler_run.diagnostic_interpolated[
                processed_cycler_run.diagnostic_interpolated.cycle_type == "hppc"]

            # Getting unique and ordered cycle index list for HPPC cycles, and choosing the hppc cycle
            hppc_cycle_list = list(set(hppc_diag_cycles.cycle_index))
            hppc_cycle_list.sort()

            # Getting unique and ordered Regular Step List (Non-unique identifier)
            reg_step_list = hppc_diag_cycles[hppc_diag_cycles.cycle_index == hppc_cycle_list[hppc_chosen]].step_index
            reg_step_list = list(set(reg_step_list))
            reg_step_list.sort()

            # The value of 1 for regular step corresponds to all of the relaxation curves in the hppc
            reg_step_relax = 1

            # Getting unique and ordered Step Counter List (unique identifier)
            step_count_list = hppc_diag_cycles[(hppc_diag_cycles.cycle_index == hppc_cycle_list[hppc_chosen]) &
                                               (hppc_diag_cycles.step_index == reg_step_list[
                                                   reg_step_relax])].step_index_counter
            step_count_list = list(set(step_count_list))
            step_count_list.sort()
            # The first one isn't a proper relaxation curve(comes out of CV) so we ignore it
            step_count_list = step_count_list[1:]
            conditions_met.append(len(step_count_list) >= n_soc_windows)

        return all(conditions_met)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Generate features listed in early prediction manuscript, primarily related to the
        so called delta Q feature
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            pd.DataFrame: features indicative of degradation, derived from the input data
        """

        rpt_dQdV_features = cls.get_rpt_dQdV_features(processed_cycler_run, diag_ref=0, diag_nr=1,
                                                                           charge_y_n=1, rpt_type='rpt_0.2C')
        hppc_features = cls.get_hppc_features(processed_cycler_run)
        relaxation_features = cls.get_all_relaxation_features(processed_cycler_run)
        fast_charge_features = cls.get_fast_charge_features(processed_cycler_run,
                                                            diagnostic_cycle_type='rpt_0.2C',
                                                            cycle_comp_num=[0, 1], Q_seg=500)
        X = pd.concat([rpt_dQdV_features, hppc_features, relaxation_features, fast_charge_features], axis=1)

        return X

    @classmethod
    def metadata_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Gather and generate information useful for filtering or subsetting the
        training feature objects for subsequent models
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            dict: information about the data source, conditions under which the run was
                performed, and other information useful for modeling and prediction
        """
        metadata = {
            'barcode': processed_cycler_run.barcode,
            'protocol': processed_cycler_run.protocol,
            'channel_id': processed_cycler_run.channel_id
        }
        return metadata

    @classmethod
    def get_rpt_dQdV_features(cls, processed_cycler_run, diag_ref=0, diag_nr=1, charge_y_n=1, rpt_type='rpt_0.2C',
                              plotting_y_n=0):
        """
        Generate features out of peakfits to rpt cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            diag_ref (int): 0 (default) reference diagnostic cycle
            diag_nr (int): 1 (default) next diagnostic cycle occurence for a specific cycle_type.
            For example, if rpt_0.2C occurs at cycle_index = [2,42,147,250.. ],
            diag_ref = 0 would correspond to cycle_index = 2
            and diag_nr = 1 would correspond to cycle_index = 42

            charge_y_n (bool): 1 = charge, 0 = discharge
            rpt_type (str): type of rpt cycle.

        Returns:
             pd.DataFrame containing features based on gaussian fits to dQdV features in rpt cycles
        """

        if ((rpt_type == 'rpt_0.2C') and (charge_y_n == 1)):
            max_nr_peaks = 4

        elif ((rpt_type == 'rpt_0.2C') and (charge_y_n == 0)):
            max_nr_peaks = 4

        elif ((rpt_type == 'rpt_1C') and (charge_y_n == 1)):
            max_nr_peaks = 4

        elif ((rpt_type == 'rpt_1C') and (charge_y_n == 0)):
            max_nr_peaks = 3

        elif ((rpt_type == 'rpt_2C') and (charge_y_n == 1)):
            max_nr_peaks = 4

        elif ((rpt_type == 'rpt_2C') and (charge_y_n == 0)):
            max_nr_peaks = 3
        else:
            raise InputError("{} is not a valid rpt cycle".format(
                rpt_type))

        peak_fit_df_ref = featurizer_helpers.generate_dQdV_peak_fits(processed_cycler_run, diag_nr=diag_ref,
                                                                     charge_y_n=charge_y_n,
                                                                     rpt_type=rpt_type, plotting_y_n=plotting_y_n,
                                                                     max_nr_peaks=max_nr_peaks)
        peak_fit_df = featurizer_helpers.generate_dQdV_peak_fits(processed_cycler_run, diag_nr=diag_nr,
                                                                 charge_y_n=charge_y_n,
                                                                 rpt_type=rpt_type, plotting_y_n=plotting_y_n,
                                                                 max_nr_peaks=max_nr_peaks)

        return 1 + (peak_fit_df - peak_fit_df_ref) / peak_fit_df_ref

    @classmethod
    def get_hppc_features(cls, processed_cycler_run, diag_pos=1, soc_window=8):
        """
        This method calculates features based on voltage, diffusion and resistance changes in hppc cycles.

        Note: Inside this function it calls function get_dr_df, but if the cell does not state of charge from 20% to
        10%, the function will fail, and throw you error messages. This will only happen after cycle 37 and on fast
        charging cells. Also, this function calls function get_v_diff, which takes in an argument soc_window, if you
        want more correlation, you should go for low state of charge, which corresponds to soc_window = 8. However,
        like the resistance feature, at cycle 142 and beyond, soc_window = 8 might fail on fast charged cells. For
        lower soc_window values, smaller than or equal to 7, this should not be a problem, but it will not give you
        high correlations.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
            if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=2 would correspond to cycle_index 142
            soc_window (int): step index counter corresponding to the soc window of interest.

        Returns:
            dataframe of features based on voltage and resistance changes over a SOC window in hppc cycles
        """
        result = pd.DataFrame()

        # diffusion features
        diffusion_features = featurizer_helpers.get_diffusion_features(processed_cycler_run, diag_pos)

        # resistance features, but the resistance feature is a 9 by 6 dataframe, so we have to make it to 1 by 54
        hppc_r = pd.DataFrame()
        # the 9 by 6 dataframe
        df_dr = featurizer_helpers.get_dr_df(processed_cycler_run, diag_pos)
        # transform this dataframe to be 1 by 54
        columns = df_dr.columns
        for column in columns:
            for r in range(len(df_dr[column])):
                name = column + str(r)
                hppc_r[name] = [df_dr[column][r]]

        # the variance of ocv features
        hppc_ocv = featurizer_helpers.get_hppc_ocv(processed_cycler_run, diag_pos)

        # the v_diff features
        v_diff = featurizer_helpers.get_v_diff(processed_cycler_run, diag_pos, soc_window)

        # merge everything together as a final result dataframe
        result = pd.concat([hppc_r, hppc_ocv, v_diff, diffusion_features], axis=1)

        return result

    @classmethod
    def get_all_relaxation_features(cls, processed_cycler_run, soc_list = np.linspace(90, 10, 9, dtype='int'),
                                    percentage_list = [50, 80, 99]):
        """
        This function returns all of the relaxation features in a panda dataframe for a given processed cycler run.

        Args:
            processed_cycler_run(beep.structure.ProcessedCyclerRun): ProcessedCyclerRun object for the cell
            you want the diagnostic features for.
            soc_list (list): list of SOCs to evaluate time constants at
            percentage_list (list): time constants to evaluate

        Returns:
            @featureDf(pd.DataFrame): Columns are either SOC{#%}_degrad{#%} where the first #% is the
            SOC % and the second #% is the time taken at what % of the final voltage value of the relaxation
            curve. The other type is names var_{#%} which is the variance of the other features taken at a
            certain % of the final voltage value of the relaxation curve.
        """

        relax_feature_array = featurizer_helpers.get_relaxation_features(processed_cycler_run)
        col_names = []
        full_feature_array = []

        for i, percentage in enumerate(percentage_list):
            col_names.append("var_{0}%".format(percentage))
            full_feature_array.append(np.var(relax_feature_array[:, i]))

            for j, soc in enumerate(soc_list):
                col_names.append("SOC{0}%_degrad{1}%".format(soc, percentage))
                full_feature_array.append(relax_feature_array[j, i])

        return pd.DataFrame(dict(zip(col_names, full_feature_array)), index=[0])


    @classmethod
    def get_fast_charge_features(cls, processed_cycler_run, diagnostic_cycle_type, cycle_comp_num=[0, 1], Q_seg=500):
        """
        Generate features listed in early prediction manuscript using both diagnostic and regular cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            diagnostic_cycle_type (str): Describes which cyle type is used for feature creation,
            options are: 'hppc', 'rpt_0.2C', 'rpt_1C', 'rpt_2C', 'reset'
            cycle_comp_num (list of two integers): contains numbers of compared cycles [0,1] for e.g.
            creates the features from the first and the second cycle of the cycle type
            Q_seg (int):  Number of cycles considered (first 500 for charging, the following 500 for discharging)


        Returns:
            X (pd.DataFrame): Dataframe containing the feature
        """

        diagnostic_interpolated = processed_cycler_run.diagnostic_interpolated

        X = pd.DataFrame(np.zeros((1, 42)))
        labels = []

        # Determine beginning and end of investigated cycle type
        index_pos_list = [i for i in range(len(diagnostic_interpolated.cycle_type))
                          if diagnostic_interpolated.cycle_type[i] == diagnostic_cycle_type]

        end_list = [index_pos_list[i] for i in range(len(index_pos_list) - 1)
                    if index_pos_list[i + 1] != index_pos_list[i] + 1]
        start_list = [index_pos_list[i] for i in range(1, len(index_pos_list))
                      if index_pos_list[i - 1] != index_pos_list[i] - 1]
        end_list.append(index_pos_list[-1])
        start_list.insert(0, index_pos_list[0])

        if diagnostic_interpolated.cycle_type[0] == diagnostic_cycle_type:
            start_list.insert(0, 1)
        if diagnostic_interpolated.cycle_type[len(diagnostic_interpolated.cycle_type) - 1] == diagnostic_cycle_type:
            end_list.append(len(diagnostic_interpolated.cycle_type) - 1)

        # Create features

        # Charging Capacity features

        # Discharging Capacity features
        Qd100_1 = diagnostic_interpolated.discharge_capacity[
                  start_list[cycle_comp_num[1]] + Q_seg + 1: start_list[cycle_comp_num[1]] + 2 * Q_seg]
        Qd10_1 = diagnostic_interpolated.discharge_capacity[
                 start_list[cycle_comp_num[0]] + Q_seg + 1: start_list[cycle_comp_num[0]] + 2 * Q_seg]
        QdDiff = [a_i - b_i for a_i, b_i in zip(Qd100_1, Qd10_1)]
        QdDiff = [elem for elem in QdDiff if (math.isnan(elem) == False)]

        X[7] = np.log10(np.absolute(np.var(QdDiff)))
        X[8] = np.log10(np.absolute(min(QdDiff)))
        X[9] = np.log10(np.absolute(np.mean(QdDiff)))
        X[10] = np.log10(np.absolute(skew(QdDiff)))
        X[11] = np.log10(np.absolute(kurtosis(QdDiff, fisher=False, bias=False)))
        X[12] = np.log10(np.sum(np.absolute(QdDiff)))
        X[13] = np.log10(np.sum(np.square(QdDiff)))

        # Charging Energy features

        # Discharging Energy features
        Ed100_1 = diagnostic_interpolated.discharge_energy[
                  start_list[cycle_comp_num[1]] + Q_seg + 1: start_list[cycle_comp_num[1]] + 2 * Q_seg]
        Ed10_1 = diagnostic_interpolated.discharge_energy[
                 start_list[cycle_comp_num[0]] + Q_seg + 1: start_list[cycle_comp_num[0]] + 2 * Q_seg]
        EdDiff = [a_i - b_i for a_i, b_i in zip(Ed100_1, Ed10_1)]
        EdDiff = [elem for elem in EdDiff if (math.isnan(elem) == False)]

        X[21] = np.log10(np.absolute(np.var(EdDiff)))
        X[22] = np.log10(np.absolute(min(EdDiff)))
        X[23] = np.log10(np.absolute(np.mean(EdDiff)))
        X[24] = np.log10(np.absolute(skew(EdDiff)))
        X[25] = np.log10(np.absolute(kurtosis(EdDiff, fisher=False, bias=False)))
        X[26] = np.log10(np.sum(np.absolute(EdDiff)))
        X[27] = np.log10(np.sum(np.square(EdDiff)))

        # Charging dQdV features

        # Discharging Capacity features
        dQdVd100_1 = diagnostic_interpolated.discharge_dQdV[
                     start_list[cycle_comp_num[1]] + Q_seg + 1: start_list[cycle_comp_num[1]] + 2 * Q_seg]
        dQdVd10_1 = diagnostic_interpolated.discharge_dQdV[
                    start_list[cycle_comp_num[0]] + Q_seg + 1: start_list[cycle_comp_num[0]] + 2 * Q_seg]
        dQdVdDiff = [a_i - b_i for a_i, b_i in zip(dQdVd100_1, dQdVd10_1)]
        dQdVdDiff = [elem for elem in dQdVdDiff if (math.isnan(elem) == False)]

        X[35] = np.log10(np.absolute(np.var(dQdVdDiff)))
        X[36] = np.log10(np.absolute(min(dQdVdDiff)))
        X[37] = np.log10(np.absolute(np.mean(dQdVdDiff)))
        X[38] = np.log10(np.absolute(skew(dQdVdDiff)))
        X[39] = np.log10(np.absolute(kurtosis(dQdVdDiff, fisher=False, bias=False)))
        X[40] = np.log10(np.sum(np.absolute(dQdVdDiff)))
        X[41] = np.log10(np.sum(np.square(dQdVdDiff)))

        operations = ['var', 'min', 'mean', 'skew', 'kurtosis', 'abs', 'square']
        quantities = ['charging_capacity', 'discharging_capacity', 'charging_energy', 'discharging_energy',
                      'charging_dQdV', 'discharging_dQdV']

        X.columns = [y + '_' + x for x in quantities for y in operations]
        return X


class DeltaQFastCharge(BeepFeatures):
    """
    Object corresponding to feature object. Includes constructors
    to create the features, object names and metadata attributes in the
    object
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    # Class name for the feature object
    class_feature_name = 'DeltaQFastCharge'

    # Class variables
    init_pred_cycle = 10
    mid_pred_cycle = 91
    final_pred_cycle = 100

    def __init__(self, name, X, metadata):
        """
        Args:
            name (str): predictor object name
            feature_object (pandas.DataFrame): features in DataFrame format.
            metadata (dict): information about the data and code used to produce features
        """
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def validate_data(cls, processed_cycler_run):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        conditions = []
        if 'cycle_index' in processed_cycler_run.summary.columns:
            conditions.append(processed_cycler_run.summary.cycle_index.max() > cls.final_pred_cycle)
            conditions.append(processed_cycler_run.summary.cycle_index.min() <= cls.init_pred_cycle)
        else:
            conditions.append(len(processed_cycler_run.summary.index) > cls.final_pred_cycle)


        return all(conditions)

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Generate features listed in early prediction manuscript, primarily related to the
        so called delta Q feature
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            pd.DataFrame: features indicative of degradation, derived from the input data
        """

        assert cls.mid_pred_cycle > 10  # Sufficient cycles for analysis
        assert cls.final_pred_cycle > cls.mid_pred_cycle  # Must have final_pred_cycle > mid_pred_cycle
        ifinal = cls.final_pred_cycle - 1  # python indexing
        imid = cls.mid_pred_cycle - 1
        iini = cls.init_pred_cycle - 1
        summary = processed_cycler_run.summary
        cycles_to_average_over = 40  # For nominal capacity, use median discharge capacity of first n cycles

        if 'step_type' in processed_cycler_run.cycles_interpolated.columns:
            interpolated_df = processed_cycler_run.cycles_interpolated[
                processed_cycler_run.cycles_interpolated.step_type == 'discharge']
        else:
            interpolated_df = processed_cycler_run.cycles_interpolated
        X = pd.DataFrame(np.zeros((1, 20)))
        labels = []
        # Discharge capacity, cycle 2 = Q(n=2)
        X[0] = summary.discharge_capacity[1]
        labels.append("discharge_capacity_cycle_2")

        # Max discharge capacity - discharge capacity, cycle 2 = max_n(Q(n)) - Q(n=2)
        X[1] = max(summary.discharge_capacity[np.arange(cls.final_pred_cycle)] - summary.discharge_capacity[1])
        labels.append("max_discharge_capacity_difference")

        # Discharge capacity, cycle 100 = Q(n=100)
        X[2] = summary.discharge_capacity[ifinal]
        labels.append("discharge_capacity_cycle_100")

        # Feature representing time-temperature integral over cycles 2 to 100
        X[3] = np.nansum(summary.time_temperature_integrated[np.arange(cls.final_pred_cycle)])
        labels.append("integrated_time_temperature_cycles_1:100")

        # Mean of charge times of first 5 cycles
        X[4] = np.nanmean(summary.charge_duration[1:6])
        labels.append("charge_time_cycles_1:5")

        # Descriptors based on capacity loss between cycles 10 and 100.
        Qd_final = interpolated_df.discharge_capacity[interpolated_df.cycle_index == ifinal]
        Qd_10 = interpolated_df.discharge_capacity[interpolated_df.cycle_index == 9]

        Vd = interpolated_df.voltage[interpolated_df.cycle_index == iini]
        Qd_diff = Qd_final.values - Qd_10.values

        # If DeltaQ(V) is not an empty array, compute summary stats, else initialize with np.nan
        # Cells discharged rapidly over a narrow voltage window run into have no interpolated discharge steps
        if len(Qd_diff):
            X[5] = np.log10(np.abs(np.nanmin(Qd_diff)))  # Minimum
            X[6] = np.log10(np.abs(np.nanmean(Qd_diff)))  # Mean
            X[7] = np.log10(np.abs(np.nanvar(Qd_diff)))  # Variance
            X[8] = np.log10(np.abs(skew(Qd_diff)))  # Skewness
            X[9] = np.log10(np.abs(kurtosis(Qd_diff)))  # Kurtosis
            X[10] = np.log10(np.abs(Qd_diff[0]))  # First difference
        else:
            X[5:11] = np.nan

        labels.append("abs_min_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_mean_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_variance_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_skew_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_kurtosis_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_first_discharge_capacity_difference_cycles_2:100")


        X[11] = max(summary.temperature_maximum[list(range(1, cls.final_pred_cycle))])  # Max T
        labels.append("max_temperature_cycles_1:100")

        X[12] = min(summary.temperature_minimum[list(range(1, cls.final_pred_cycle))])  # Min T
        labels.append("min_temperature_cycles_1:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 2 to 100

        X[13], X[14] = np.polyfit(
            list(range(1, cls.final_pred_cycle)),
            summary.discharge_capacity[list(range(1, cls.final_pred_cycle))], 1)

        labels.append("slope_discharge_capacity_cycle_number_2:100")
        labels.append("intercept_discharge_capacity_cycle_number_2:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 91 to 100
        X[15], X[16] = np.polyfit(
            list(range(imid, cls.final_pred_cycle)),
            summary.discharge_capacity[list(range(imid, cls.final_pred_cycle))], 1)
        labels.append("slope_discharge_capacity_cycle_number_91:100")
        labels.append("intercept_discharge_capacity_cycle_number_91:100")

        IR_trend = summary.dc_internal_resistance[list(range(1, cls.final_pred_cycle))]
        if any(v == 0 for v in IR_trend):
            IR_trend[IR_trend == 0] = np.nan

        # Internal resistance minimum
        X[17] = np.nanmin(IR_trend)
        labels.append("min_internal_resistance_cycles_2:100")

        # Internal resistance at cycle 2
        X[18] = summary.dc_internal_resistance[1]
        labels.append("internal_resistance_cycle_2")

        # Internal resistance at cycle 100 - cycle 2
        X[19] = summary.dc_internal_resistance[ifinal] - summary.dc_internal_resistance[1]
        labels.append("internal_resistance_difference_cycles_2:100")

        # Nominal capacity
        X[20] = np.median(summary.discharge_capacity.iloc[0:cycles_to_average_over])
        labels.append("nominal_capacity_by_median")

        X.columns = labels
        return X

    @classmethod
    def metadata_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Gather and generate information useful for filtering or subsetting the
        training feature objects for subsequent models
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            dict: information about the data source, conditions under which the run was
                performed, and other information useful for modeling and prediction
        """
        metadata = {
            'barcode': processed_cycler_run.barcode,
            'protocol': processed_cycler_run.protocol,
            'channel_id': processed_cycler_run.channel_id
        }
        return metadata


class TrajectoryFastCharge(DeltaQFastCharge):
    """
    Object corresponding to cycle numbers at which the capacity drops below
     specific percentages of the initial capacity. Computed on the discharge
     portion of the regular fast charge cycles.

        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    # Class name for the feature object
    class_feature_name = 'TrajectoryFastCharge'

    def __init__(self, name, X, metadata):
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Calculate the outcomes from the input data. In particular, the number of cycles
        where we expect to reach certain thresholds of capacity loss
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            pd.DataFrame: cycles at which capacity/energy degradation exceeds thresholds
        """
        y = processed_cycler_run.cycles_to_reach_set_capacities(
            thresh_max_cap=0.98, thresh_min_cap=0.78, interval_cap=0.03)
        return y

class DiagnosticProperties(DiagnosticCyclesFeatures):
    """
    This class stores fractional levels of degradation in discharge capacity and discharge energy
    relative to the first cycle at each diagnostic cycle, grouped by diagnostic cycle type.

        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    # Class name for the feature object
    class_feature_name = 'DiagnosticProperties'

    def __init__(self, name, X, metadata):
        super().__init__(name, X, metadata)
        self.name = name
        self.X = X
        self.metadata = metadata

    @classmethod
    def features_from_processed_cycler_run(cls, processed_cycler_run):
        """
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
        Returns:
            pd.DataFrame: cycles at which capacity/energy degradation exceeds thresholds
        """

        quantities = ['discharge_energy', 'discharge_capacity']
        cycle_types = processed_cycler_run.diagnostic_summary.cycle_type.unique()
        X = pd.DataFrame()
        for quantity in quantities:
            for cycle_type in cycle_types:
                summary_diag_cycle_type = DiagnosticProperties.get_fractional_quantity_remaining(processed_cycler_run,
                                                                                                 quantity, cycle_type)
                summary_diag_cycle_type['cycle_type'] = cycle_type
                summary_diag_cycle_type['metric'] = quantity
                X = X.append(summary_diag_cycle_type)

        return X

    @staticmethod
    def get_fractional_quantity_remaining(processed_cycler_run, metric='discharge_energy',
                                          diagnostic_cycle_type='rpt_0.2C'):
        """
        Determine relative loss of <metric> in diagnostic_cycles of type <diagnostic_cycle_type> after 100 regular cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): information about cycler run
            metric (str): column name to use for measuring degradation
            diagnostic_cycle_type (str): the diagnostic cycle to use for computing the amount of degradation

        Returns:
            a dataframe with cycle_index and corresponding degradation relative to the first measured value
        """
        summary_diag_cycle_type = processed_cycler_run.diagnostic_summary[
            (processed_cycler_run.diagnostic_summary.cycle_type == diagnostic_cycle_type)
            & (processed_cycler_run.diagnostic_summary.cycle_index > 100)].reset_index()
        summary_diag_cycle_type = summary_diag_cycle_type[['cycle_index', metric]]
        summary_diag_cycle_type[metric] = summary_diag_cycle_type[metric] / \
                                          processed_cycler_run.diagnostic_summary[metric].iloc[0]
        summary_diag_cycle_type.columns = ['cycle_index', 'fractional_metric']
        return summary_diag_cycle_type


class DegradationPredictor(MSONable):
    """
    Object corresponding to feature matrix. Includes constructors
    to initialize the feature vectors.
    Attributes:
        name (str): predictor object name.
        X (pandas.DataFrame): data as records x features.
        y (pandas.DataFrame): targets.
        feature_labels (list): feature labels.
        predict_only (bool): True/False to specify predict/train mode.
        prediction_type (str): Type of regression - 'single' vs 'multi'.
        predicted_quantity (str): 'cycle' or 'capacity'.
        nominal_capacity (float):
    """

    def __init__(self, name, X, feature_labels=None, y=None, nominal_capacity=1.1,
                 predict_only=False, predicted_quantity="cycle", prediction_type="multi"):
        """
        Args:
            name (str): predictor object name
            X (pandas.DataFrame): features in DataFrame format.
            name (str): name of method for featurization.
            y (pandas.Dataframe or float): one or more outcomes.
            predict_only (bool): True/False to specify predict/train mode.
            predicted_quantity (str): 'cycle' or 'capacity'.
            prediction_type (str): Type of regression - 'single' vs 'multi'.
        """
        self.name = name
        self.X = X
        self.feature_labels = feature_labels
        self.predict_only = predict_only
        self.prediction_type = prediction_type
        self.predicted_quantity = predicted_quantity
        self.y = y
        self.nominal_capacity = nominal_capacity

    @classmethod
    def from_processed_cycler_run_file(cls, path, features_label='full_model', predict_only=False,
                                       predicted_quantity='cycle', prediction_type='multi',
                                       diagnostic_features=False):
        """
        Args:
            path (str): string corresponding to file path with ProcessedCyclerRun object.
            features_label (str): name of method for featurization.
            predict_only (bool): True/False to specify predict/train mode.
            predicted_quantity (str): 'cycle' or 'capacity'.
            prediction_type (str): Type of regression - 'single' vs 'multi'.
            diagnostic_features (bool): whether to compute diagnostic features.
        """
        processed_cycler_run = loadfn(path)

        if features_label == 'full_model':
            return cls.init_full_model(processed_cycler_run, predict_only=predict_only,
                                       predicted_quantity=predicted_quantity,
                                       diagnostic_features=diagnostic_features,
                                       prediction_type=prediction_type)
        else:
            raise NotImplementedError

    @classmethod
    def init_full_model(cls, processed_cycler_run, init_pred_cycle=10, mid_pred_cycle=91,
                        final_pred_cycle=100, predict_only=False, prediction_type='multi',
                        predicted_quantity="cycle", diagnostic_features=False):
        """
        Generate features listed in early prediction manuscript
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): information about cycler run
            init_pred_cycle (int): index of initial cycle index used for predictions
            mid_pred_cycle (int): index of intermediate cycle index used for predictions
            final_pred_cycle (int): index of highest cycle index used for predictions
            predict_only (bool): whether or not to include cycler life in the object
            prediction_type (str): 'single': cycle life to reach 80% capacity.
                                   'multi': remaining capacity at fixed cycles
            predicted_quantity (str): quantity being predicted - cycles/capacity
            diagnostic_features (bool): whether or not to compute diagnostic features
        Returns:
            beep.featurize.DegradationPredictor: DegradationPredictor corresponding to the ProcessedCyclerRun file.
        """
        assert mid_pred_cycle > 10, 'Insufficient cycles for analysis'
        assert final_pred_cycle > mid_pred_cycle, 'Must have final_pred_cycle > mid_pred_cycle'
        ifinal = final_pred_cycle - 1  # python indexing
        imid = mid_pred_cycle - 1
        iini = init_pred_cycle - 1
        summary = processed_cycler_run.summary
        assert len(processed_cycler_run.summary) > final_pred_cycle, 'cycle count must exceed final_pred_cycle'
        cycles_to_average_over = 40  # For nominal capacity, use median discharge capacity of first n cycles

        # Features in "nature energy" set only use discharge portion of the cycle
        if 'step_type' in processed_cycler_run.cycles_interpolated.columns:
            interpolated_df = processed_cycler_run.cycles_interpolated[
                processed_cycler_run.cycles_interpolated.step_type == 'discharge']
        else:
            interpolated_df = processed_cycler_run.cycles_interpolated

        X = pd.DataFrame(np.zeros((1, 20)))
        labels = []
        # Discharge capacity, cycle 2 = Q(n=2)
        X[0] = summary.discharge_capacity[1]
        labels.append("discharge_capacity_cycle_2")

        # Max discharge capacity - discharge capacity, cycle 2 = max_n(Q(n)) - Q(n=2)
        X[1] = max(summary.discharge_capacity[np.arange(final_pred_cycle)] - summary.discharge_capacity[1])
        labels.append("max_discharge_capacity_difference")

        # Discharge capacity, cycle 100 = Q(n=100)
        X[2] = summary.discharge_capacity[ifinal]
        labels.append("discharge_capacity_cycle_100")

        # Feature representing time-temperature integral over cycles 2 to 100
        X[3] = np.nansum(summary.time_temperature_integrated[np.arange(final_pred_cycle)])
        labels.append("integrated_time_temperature_cycles_1:100")

        # Mean of charge times of first 5 cycles
        X[4] = np.nanmean(summary.charge_duration[1:6])
        labels.append("charge_time_cycles_1:5")

        # Descriptors based on capacity loss between cycles 10 and 100.
        Qd_final = interpolated_df.discharge_capacity[interpolated_df.cycle_index == ifinal]
        Qd_10 = interpolated_df.discharge_capacity[interpolated_df.cycle_index == 9]

        Vd = interpolated_df.voltage[interpolated_df.cycle_index == iini]
        Qd_diff = Qd_final.values - Qd_10.values

        X[5] = np.log10(np.abs(np.min(Qd_diff)))  # Minimum
        labels.append("abs_min_discharge_capacity_difference_cycles_2:100")

        X[6] = np.log10(np.abs(np.mean(Qd_diff)))  # Mean
        labels.append("abs_mean_discharge_capacity_difference_cycles_2:100")

        X[7] = np.log10(np.abs(np.var(Qd_diff)))  # Variance
        labels.append("abs_variance_discharge_capacity_difference_cycles_2:100")

        X[8] = np.log10(np.abs(skew(Qd_diff)))  # Skewness
        labels.append("abs_skew_discharge_capacity_difference_cycles_2:100")

        X[9] = np.log10(np.abs(kurtosis(Qd_diff)))  # Kurtosis
        labels.append("abs_kurtosis_discharge_capacity_difference_cycles_2:100")

        X[10] = np.log10(np.abs(Qd_diff[0]))  # First difference
        labels.append("abs_first_discharge_capacity_difference_cycles_2:100")

        X[11] = max(summary.temperature_maximum[list(range(1, final_pred_cycle))])  # Max T
        labels.append("max_temperature_cycles_1:100")

        X[12] = min(summary.temperature_minimum[list(range(1, final_pred_cycle))])  # Min T
        labels.append("min_temperature_cycles_1:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 2 to 100

        X[13], X[14] = np.polyfit(
            list(range(1, final_pred_cycle)),
            summary.discharge_capacity[list(range(1, final_pred_cycle))], 1)

        labels.append("slope_discharge_capacity_cycle_number_2:100")
        labels.append("intercept_discharge_capacity_cycle_number_2:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 91 to 100
        X[15], X[16] = np.polyfit(
            list(range(imid, final_pred_cycle)),
            summary.discharge_capacity[list(range(imid, final_pred_cycle))], 1)
        labels.append("slope_discharge_capacity_cycle_number_91:100")
        labels.append("intercept_discharge_capacity_cycle_number_91:100")

        IR_trend = summary.dc_internal_resistance[list(range(1, final_pred_cycle))]
        if any(v == 0 for v in IR_trend):
            IR_trend[IR_trend == 0] = np.nan

        # Internal resistance minimum
        X[17] = np.nanmin(IR_trend)
        labels.append("min_internal_resistance_cycles_2:100")

        # Internal resistance at cycle 2
        X[18] = summary.dc_internal_resistance[1]
        labels.append("internal_resistance_cycle_2")

        # Internal resistance at cycle 100 - cycle 2
        X[19] = summary.dc_internal_resistance[ifinal] - summary.dc_internal_resistance[1]
        labels.append("internal_resistance_difference_cycles_2:100")

        X.columns = labels
        if predict_only:
            y = None
        else:
            if prediction_type == 'single':
                y = processed_cycler_run.get_cycle_life()
            elif prediction_type == 'multi':
                if predicted_quantity == 'cycle':
                    y = processed_cycler_run.cycles_to_reach_set_capacities(
                        thresh_max_cap=0.98, thresh_min_cap=0.78, interval_cap=0.03)
                elif predicted_quantity == 'capacity':
                    y = processed_cycler_run.capacities_at_set_cycles()
                else:
                    raise NotImplementedError(
                        "{} predicted_quantity type not implemented".format(
                            predicted_quantity))
        nominal_capacity = np.median(summary.discharge_capacity.iloc[0:cycles_to_average_over])

        return cls('full_model', X, feature_labels=labels, y=y,
                   nominal_capacity=nominal_capacity, predict_only=predict_only,
                   prediction_type=prediction_type, predicted_quantity=predicted_quantity)

    def as_dict(self):
        """
        Method for dictionary serialization
        Returns:
            dict: corresponding to dictionary for serialization
        """
        obj = {"@module": self.__class__.__module__,
               "@class": self.__class__.__name__,
               "name": self.name,
               "X": self.X.to_dict("list"),
               "feature_labels": self.feature_labels,
               "predict_only": self.predict_only,
               "prediction_type": self.prediction_type,
               "nominal_capacity": self.nominal_capacity
               }
        if isinstance(self.y, pd.DataFrame):
            obj["y"] = self.y.to_dict("list")
        else:
            obj["y"] = self.y
        return obj

    @classmethod
    def from_dict(cls, d):
        """MSONable deserialization method"""
        d['X'] = pd.DataFrame(d['X'])
        return cls(**d)


def add_file_prefix_to_path(path, prefix):
    """
    Helper function to add file prefix to path.

    Args:
        path (str): full path to file.
        prefix (str): prefix for file.

    Returns:
        str: path with prefix appended to filename.

    """
    split_path = list(os.path.split(path))
    split_path[-1] = prefix + split_path[-1]
    return os.path.join(*split_path)


def process_file_list_from_json(file_list_json, processed_dir='data-share/features/'):
    """
    Function to take a json file containing processed cycler run file locations,
    extract features, dump the processed file into a predetermined directory,
    and return a jsonable dict of feature file locations.

    Args:
        file_list_json (str): json string or json filename corresponding
            to a dictionary with a file_list attribute,
            if this string ends with ".json", a json file is assumed
            and loaded, otherwise interpreted as a json string.
        processed_dir (str): location for processed cycler run output files
            to be placed.
        features_label (str): name of feature generation method.
        predict_only (bool): whether to calculate predictions or not.
        prediction_type (str): Single or multi-point predictions.
        predicted_quantity (str): quantity being predicted - cycle or capacity.

    Returns:
        str: json string of feature files (with key "file_list").

    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup Events
    events = KinesisEvents(service='DataAnalyzer', mode=file_list_data['mode'])

    # Add root path to processed_dir
    processed_dir = os.path.join(os.environ.get("BEEP_PROCESSING_DIR", "/"),
                                 processed_dir)
    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)

    file_list = file_list_data['file_list']
    run_ids = file_list_data['run_list']
    processed_run_list = []
    processed_result_list = []
    processed_message_list = []
    processed_paths_list = []

    for path, run_id in zip(file_list, run_ids):
        logger.info('run_id=%s featurizing=%s', str(run_id), path, extra=s)
        processed_cycler_run = loadfn(path)

        featurizer_classes = [DeltaQFastCharge, TrajectoryFastCharge, DiagnosticCyclesFeatures, DiagnosticProperties]
        for featurizer_class in featurizer_classes:
            featurizer = featurizer_class.from_run(path, processed_dir, processed_cycler_run)
            if featurizer:
                dumpfn(featurizer, featurizer.name)
                processed_paths_list.append(featurizer.name)
                processed_run_list.append(run_id)
                processed_result_list.append("success")
                processed_message_list.append({'comment': '',
                                               'error': ''})
                logger.info('Successfully generated %s', featurizer.name, extra=s)
            else:
                processed_paths_list.append(path)
                processed_run_list.append(run_id)
                processed_result_list.append("incomplete")
                processed_message_list.append({'comment': 'Insufficient or incorrect data for featurization',
                                               'error': ''})
                logger.info('Unable to featurize %s', path, extra=s)

    output_data = {"file_list": processed_paths_list,
                   "run_list": processed_run_list,
                   "result_list": processed_result_list,
                   "message_list": processed_message_list
                   }

    events.put_analyzing_event(output_data, 'featurizing', 'complete')
    # Return jsonable file list
    return json.dumps(output_data)


def main():
    """
    Main function of this module, takes in arguments of an input
    and output filename corresponding to structured cycler run data
    and creates a predictor object output for analysis/ML processing

    Returns:
        None

    """
    # Parse args and construct initial cycler run
    logger.info('starting', extra=s)
    logger.info('Running version=%s', __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args['INPUT_JSON']
        print(process_file_list_from_json(input_json), end="")
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info('finish', extra=s)
    return None


if __name__ == "__main__":
    main()



# Copyright 2019 Toyota Research Institute. All rights reserved.
"""
Module and script for processing cycler run CSVs into structured data
for featurization and analysis.  Contains RawCyclerRun and ProcessedCyclerRun
objects, which parse and structure data, along with utility
functions for manipulating tabular data.

Usage:
    structure [INPUT_JSON]

Options:
    -h --help       Show this screen
    --version       Show version


The `structure` script will run the data structuring on specified filenames corresponding
to validated raw cycler files.  It places the structured datafiles in `/data-share/structure`.

The input json must contain the following fields:
* `file_list` - a list of full path filenames which have been processed
* `validity` - a list of boolean validation results, e. g. `[True, True, False]`

The output json contains the following fields:

* `invalid_file_list` - a list of invalid files according to the validity
* `file_list` - a list of files which have been structured into processed_cycler_runs

Example:
```angular2
$ structure '{"validity": [false, false, true],
             file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
                          "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv",
                          "/data-share/renamed_cycler_files/FastCharge/FastCharge_2_CH29.csv"]}''
{"invalid_file_list": ["/data-share/renamed_cycler_files/FastCharge/FastCharge_0_CH33.csv",
                       "/data-share/renamed_cycler_files/FastCharge/FastCharge_1_CH44.csv"],
 "file_list": ["/data-share/structure/FastCharge_2_CH29_structure.json"]}
```
"""

import json
import re
from datetime import datetime

import pandas as pd
import numpy as np
import os
import pytz
import time
from scipy import integrate
import itertools

from monty.json import MSONable
from docopt import docopt
from monty.serialization import loadfn, dumpfn
from glob import glob
from beep import tqdm

from beep import StringIO, MODULE_DIR, ENVIRONMENT
from beep.validate import ValidatorBeep, BeepValidationError
from beep.collate import add_suffix_to_filename
from beep.conversion_schemas import ARBIN_CONFIG, MACCOR_CONFIG, \
    FastCharge_CONFIG, xTesladiag_CONFIG, INDIGO_CONFIG, BIOLOGIC_CONFIG, \
    STRUCTURE_DTYPES
from beep.utils import KinesisEvents
from beep import logger, __version__

s = {'service': 'DataStructurer'}


class RawCyclerRun(MSONable):
    """
    Object corresponding to parsed cycler run, includes a factory method to
    construct object from a file path (e. g. "test_file.csv")

    Attributes:
        data (pandas.DataFrame): DataFrame corresponding to cycler run data.
        metadata (dict): Dict corresponding to cycler run metadata.
        eis (beep.structure.EISpectrum): electrochemical impedence
            spectrum object. Defaults to None.
        validate (bool): whether or not to validate DataFrame upon
            instantiation. Defaults to None.
    """
    # These define float and int columns for numeric binary save files
    FLOAT_COLUMNS = ["test_time", "current", "voltage", "charge_capacity",
                     "discharge_capacity",
                     "charge_energy", "discharge_energy",
                     "internal_resistance", "temperature"]
    INT_COLUMNS = ["step_index", "cycle_index"]

    def __init__(self, data, metadata, eis=None, validate=False, filename=None):
        """
        Args:
            data (pandas.DataFrame): DataFrame corresponding to cycler run data
            metadata (dict): Dict corresponding to cycler run metadata
            eis (beep.structure.EISpectrum): electrochemical impedence
                spectrum object. Defaults to None
            validate (bool): whether or not to validate DataFrame upon
                instantiation. Defaults to None.
        """
        if validate:
            validator = ValidatorBeep()
            is_valid = validator.validate_arbin_dataframe(data)
            if not is_valid:
                raise BeepValidationError("Beep validation failed")

        self.data = data
        self.metadata = metadata
        self.eis = eis
        self.filename = filename

    @classmethod
    def from_file(cls, path, validate=False):
        """
        Factory method to invoke RawCyclerRun from filename with recognition of
        type from filename, using corresponding class method as constructor.

        Args:
            path (str): string corresponding to file path.
            validate (bool): whether or not to validate file.

        Returns:
            beep.structure.RawCyclerRun: RawCyclerRun corresponding to parsed file(s).

        """
        if re.match(ARBIN_CONFIG['file_pattern'], path):
            return cls.from_arbin_file(path, validate)

        elif re.match(MACCOR_CONFIG['file_pattern'], path):
            return cls.from_maccor_file(path, False, validate)

        elif re.match(INDIGO_CONFIG['file_pattern'], path):
            return cls.from_indigo_file(path, validate)

        elif re.match(BIOLOGIC_CONFIG['file_pattern'], path):
            return cls.from_biologic_file(path, validate)

        else:
            raise ValueError("{} does not match any known file pattern".format(path))

    def get_interpolated_steps(self, v_range, resolution, step_type='discharge', reg_cycles=None, axis='voltage'):
        """
        Gets interpolated cycles for the step specified, charge or discharge.

        Args:
            v_range ([Float, Float]): list of two floats that define
                the voltage interpolation range endpoints.
            resolution (int): resolution of interpolated data.
            step_type (str): which step to interpolate i.e. 'charge' or 'discharge'
            reg_cycles (list): list containing cycle indicies of regular cycles
            axis (str): which column to use for interpolation

        Returns:
            pandas.DataFrame: DataFrame corresponding to interpolated values.
        """
        if step_type is 'discharge':
            step_filter = determine_whether_step_is_discharging
        elif step_type is 'charge':
            step_filter = determine_whether_step_is_charging
        else:
            raise ValueError("{} is not a recognized step type")
        incl_columns = ["voltage", "current", "charge_capacity", "discharge_capacity",
                        "internal_resistance", "temperature"]
        all_dfs = []
        cycle_indices = self.data.cycle_index.unique()
        cycle_indices = [c for c in cycle_indices if c in reg_cycles]
        cycle_indices.sort()
        for cycle_index in tqdm(cycle_indices):
            # Use a cycle_index mask instead of a global groupby to save memoryÍ
            new_df = self.data.loc[self.data["cycle_index"] == cycle_index].groupby("step_index").filter(step_filter)
            if new_df.size == 0:
                continue
            if axis in ['charge_capacity', 'discharge_capacity']:
                axis_range = [self.data[axis].min(), self.data[axis].max()]
                new_df = get_interpolated_data(new_df, axis, field_range=axis_range,
                                               columns=incl_columns, resolution=resolution)
            elif axis == 'test_time':
                axis_range = [new_df[axis].min(), new_df[axis].max()]
                new_df = get_interpolated_data(new_df, axis, field_range=axis_range,
                                               columns=incl_columns, resolution=resolution)
            elif axis == 'voltage':
                new_df = get_interpolated_data(new_df, axis, field_range=v_range,
                                               columns=incl_columns, resolution=resolution)
            else:
                raise NotImplementedError
            new_df['cycle_index'] = cycle_index
            new_df['step_type'] = step_type
            new_df['step_type'] = new_df['step_type'].astype('category')
            all_dfs.append(new_df)

        # Ignore the index to avoid issues with overlapping voltages
        result = pd.concat(all_dfs, ignore_index=True)

        # Cycle_index gets a little weird about typing, so round it here
        result.cycle_index = result.cycle_index.round()

        return result

    def get_interpolated_cycles(self, v_range=None, resolution=1000, diagnostic_available=None):
        """
        Gets interpolated cycles for both charge and discharge steps.

        Args:
            v_range ([Float, Float]): list of two floats that define
                the voltage interpolation range endpoints.
            resolution (int): resolution of interpolated data.
            diagnostic_available (dict): dictionary containing information about
                location of diagnostic cycles

        Returns:
            pandas.DataFrame: DataFrame corresponding to interpolated values.
        """
        if diagnostic_available:
            diag_cycles = list(itertools.chain.from_iterable(
                [list(range(i, i + diagnostic_available['length'])) for i in
                 diagnostic_available['diagnostic_starts_at']
                 if i <= self.data.cycle_index.max()]))
            reg_cycles = [i for i in self.data.cycle_index.unique() if i not in diag_cycles]
        else:
            reg_cycles = [i for i in self.data.cycle_index.unique()]

        v_range = v_range or [2.8, 3.5]
        interpolated_discharge = self.get_interpolated_steps(v_range,
                                                             resolution,
                                                             step_type='discharge',
                                                             reg_cycles=reg_cycles,
                                                             axis='voltage')
        interpolated_charge = self.get_interpolated_steps(v_range,
                                                          resolution,
                                                          step_type='charge',
                                                          reg_cycles=reg_cycles,
                                                          axis='charge_capacity')
        result = pd.concat([interpolated_discharge, interpolated_charge], ignore_index=True)
        result = result.astype(STRUCTURE_DTYPES['cycles_interpolated'])

        return result

    def as_dict(self):
        """
        Method for dictionary/json serialization hook in MSONable

        Returns:
            dict: Representation as dictionary.
        """
        obj = {"@module": self.__class__.__module__,
               "@class": self.__class__.__name__,
               "data": self.data.to_dict('list'),
               "metadata": self.metadata,
               "eis": self.eis}
        return obj

    @classmethod
    def from_dict(cls, d):
        """
        Method for dictionary/json deserialization hook in MSONable

        Returns:
            beep.structure.RawCyclerRun:
        """

        data = pd.DataFrame(d['data'])
        data = data.sort_index()
        return cls(data, d['metadata'], d['eis'])

    def get_summary(self, diagnostic_available=None, nominal_capacity=1.1,
                    full_fast_charge=0.8, cycle_complete_discharge_ratio=0.97,
                    cycle_complete_vmin=3.3, cycle_complete_vmax=3.3):
        """
        Gets summary statistics for data according to cycle number. Summary data
        must be float or int type for compatibility with other methods

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            cycle_complete_discharge_ratio (float): expected ratio
                discharge/charge at the end of any complete cycle
            cycle_complete_vmin (float): expected voltage minimum achieved
                in any complete cycle
            cycle_complete_vmax (float): expected voltage maximum achieved
                in any complete cycle

        Returns:
            pandas.DataFrame: summary statistics by cycle.

        """
        #Filter out only regular cycles for summary stats. Diagnostic summary computed separately
        if diagnostic_available:
            diag_cycles = list(itertools.chain.from_iterable(
                [list(range(i, i + diagnostic_available['length'])) for i in
                 diagnostic_available['diagnostic_starts_at']
                 if i <= self.data.cycle_index.max()]))
            reg_cycles_at = [i for i in self.data.cycle_index.unique() if i not in diag_cycles]
        else:
            reg_cycles_at = [i for i in self.data.cycle_index.unique()]

        summary = self.data.groupby("cycle_index").agg({
            "cycle_index": "first",
            "discharge_capacity": "max",
            "charge_capacity": "max",
            "discharge_energy": "max",
            "charge_energy": "max",
            "internal_resistance": "last",
            "temperature": ["max", "mean", "min"],
            "date_time_iso": "first"})

        summary.columns = ['cycle_index', 'discharge_capacity', 'charge_capacity',
                           'discharge_energy', 'charge_energy',
                           'dc_internal_resistance', 'temperature_maximum',
                           'temperature_average', 'temperature_minimum',
                           'date_time_iso']
        summary = summary[summary.index.isin(reg_cycles_at)]
        summary['energy_efficiency'] = summary['discharge_energy']/summary['charge_energy']
        summary.loc[~np.isfinite(summary['energy_efficiency']), 'energy_efficiency'] = np.NaN
        summary['charge_throughput'] = summary.charge_capacity.cumsum()
        summary['energy_throughput'] = summary.charge_energy.cumsum()

        # This method for computing charge start and end times implicitly
        # assumes that a cycle starts with a charge step and is then followed
        # by discharge step.
        charge_start_time = self.data.groupby('cycle_index', as_index=False)['date_time_iso'].agg('first')
        charge_finish_time = self.data[self.data.charge_capacity >= nominal_capacity*full_fast_charge].\
            groupby('cycle_index', as_index=False)['date_time_iso'].agg('first')

        # Left merge, since some cells might not reach desired levels of
        # charge_capacity and will have NaN for charge duration
        merged = charge_start_time.merge(charge_finish_time, on='cycle_index', how='left')

        # Charge duration stored in seconds - note that date_time_iso is only ~1sec resolution
        time_diff = np.subtract(pd.to_datetime(merged.date_time_iso_y, utc=True, errors='coerce'),
                                pd.to_datetime(merged.date_time_iso_x, errors='coerce'))
        summary["charge_duration"] = np.round(time_diff/np.timedelta64(1, 's'), 2)

        # Compute time since start of cycle in minutes. This comes handy
        # for featurizing time-temperature integral
        self.data['time_since_cycle_start'] = pd.to_datetime(self.data['date_time_iso']) - \
            pd.to_datetime(self.data.groupby('cycle_index')['date_time_iso'].transform('first'))
        self.data['time_since_cycle_start'] = (self.data['time_since_cycle_start'] / np.timedelta64(1, 's'))/60

        # Group by cycle index and integrate time-temperature
        # using a lambda function.
        summary['time_temperature_integrated'] = \
            self.data.groupby('cycle_index').apply(lambda g: integrate.trapz(g.temperature, x=g.time_since_cycle_start))

        # Drop the time since cycle start column
        self.data.drop(columns=['time_since_cycle_start'])

        # Determine if any of the cycles has been paused
        summary['paused'] = self.data.groupby("cycle_index").apply(determine_paused)

        summary = summary.astype(STRUCTURE_DTYPES['summary'])

        last_voltage = self.data.loc[self.data['cycle_index'] == self.data['cycle_index'].max()]['voltage']
        if ((last_voltage.min() < cycle_complete_vmin) and (last_voltage.max() > cycle_complete_vmax) and
            ((summary.iloc[[-1]])['discharge_capacity'].iloc[0] > cycle_complete_discharge_ratio
             * (summary.iloc[[-1]])['charge_capacity'].iloc[0])):
            return summary
        else:
            return summary.iloc[:-1]

    def get_diagnostic_summary(self, diagnostic_available):
        """
        Gets summary statistics for data according to location of
        diagnostic cycles in the data

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
                as list, 'length' of the diagnostic in cycles and location
                of the diagnostic by cycle index

        Returns:
            (DataFrame) of summary statistics by cycle

        """

        max_cycle = self.data.cycle_index.max()
        starts_at = [i for i in diagnostic_available['diagnostic_starts_at']
                     if i <= max_cycle]
        diag_cycles_at = list(itertools.chain.from_iterable(
            [list(range(i, i + diagnostic_available['length'])) for i in starts_at]))
        diag_summary = self.data.groupby("cycle_index").agg({
            "discharge_capacity": "max",
            "charge_capacity": "max",
            "discharge_energy": "max",
            "charge_energy": "max",
            "temperature": ["max", "mean", "min"],
            "date_time_iso": "first",
            "cycle_index": "first"},
        )

        diag_summary.columns = ['discharge_capacity', 'charge_capacity',
                                'discharge_energy', 'charge_energy',
                                'temperature_maximum', 'temperature_average',
                                'temperature_minimum', 'date_time_iso',
                                'cycle_index']
        diag_summary = diag_summary[diag_summary.index.isin(diag_cycles_at)]

        diag_summary['coulombic_efficiency'] = diag_summary['discharge_capacity'] \
                                               / diag_summary['charge_capacity']
        diag_summary['paused'] = self.data.groupby("cycle_index").apply(determine_paused)

        diag_summary.reset_index(drop=True, inplace=True)

        diag_summary['cycle_type'] = pd.Series(diagnostic_available['cycle_type'] * len(starts_at))

        diag_summary = diag_summary.astype(STRUCTURE_DTYPES['diagnostic_summary'])

        return diag_summary

    def get_interpolated_diagnostic_cycles(self, diagnostic_available,
                                           resolution=1000, v_resolution=0.0005):
        """
        Interpolates data according to location and type of diagnostic
        cycles in the data

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
                as list, 'length' of the diagnostic in cycles and location
                of the diagnostic
            resolution (int): resolution of interpolation
            v_resolution (int): voltage delta to set for range based interpolation

        Returns:
            (DataFrame) of interpolated diagnostic steps by step and cycle

        """
        # Get the project name and the parameter file for the diagnostic
        project_name_list = get_project_sequence(self.filename)
        diag_path = os.path.join(MODULE_DIR, 'procedure_templates')
        v_range = get_diagnostic_parameters(
            diagnostic_available, diag_path, project_name_list[0])

        # Determine the cycles and types of the diagnostic cycles
        max_cycle = self.data.cycle_index.max()
        starts_at = [i for i in diagnostic_available['diagnostic_starts_at']
                     if i <= max_cycle]
        diag_cycles_at = list(itertools.chain.from_iterable(
            [range(i, i + diagnostic_available['length']) for i in starts_at]))
        # Duplicate cycle type list end to end for each starting index
        diag_cycle_type = diagnostic_available['cycle_type'] * len(starts_at)
        if not len(diag_cycles_at) == len(diag_cycle_type):
            errmsg = "Diagnostic cycles, {}, and diagnostic cycle types, "\
                     "{}, are unequal lengths".format(diag_cycles_at, diag_cycle_type)
            raise ValueError(errmsg)

        diag_data = self.data[self.data['cycle_index'].isin(diag_cycles_at)]

        # Convert date_time_iso field into pd.datetime object
        diag_data.loc[:, 'date_time_iso'] = pd.to_datetime(diag_data['date_time_iso'])

        # Convert datetime into seconds to allow interpolation of time
        diag_data.loc[:, 'datetime_seconds'] = [time.mktime(t.timetuple())
                                                if t is not pd.NaT else float('nan')
                                                for t in diag_data['date_time_iso']]

        # Counter to ensure non-contiguous repeats of step_index
        # within same cycle_index are grouped separately
        diag_data.loc[:, 'step_index_counter'] = 0

        for cycle_index in diag_cycles_at:
            indices = diag_data.loc[diag_data.cycle_index == cycle_index].index
            step_index_list = diag_data.step_index.loc[indices]
            diag_data.loc[indices, 'step_index_counter'] = \
                step_index_list.ne(step_index_list.shift()).cumsum()

        group = diag_data.groupby(["cycle_index", "step_index", "step_index_counter"])
        incl_columns = ["current", "charge_capacity", "discharge_capacity",
                        "charge_energy", "discharge_energy", "internal_resistance",
                        "temperature", "datetime_seconds", "test_time"]

        diag_dict = {}
        for cycle in diag_data.cycle_index.unique():
            diag_dict.update({cycle: None})
            steps = diag_data[diag_data.cycle_index == cycle].step_index.unique()
            diag_dict[cycle] = list(steps)

        all_dfs = []
        for (cycle_index, step_index, step_index_counter), df in tqdm(group):
            if diag_cycle_type[diag_cycles_at.index(cycle_index)] == 'hppc':
                v_hppc_step = [df.voltage.min(), df.voltage.max()]
                hppc_resolution = int((df.voltage.max() - df.voltage.min()) / v_resolution)
                new_df = get_interpolated_data(df, field_name="voltage", field_range=v_hppc_step,
                                               columns=incl_columns, resolution=hppc_resolution)
            else:
                new_df = get_interpolated_data(df, field_name="voltage", field_range=v_range,
                                               columns=incl_columns, resolution=resolution)

            #Convert interpolated time in seconds back to datetime
            new_df['date_time_iso'] = [datetime.utcfromtimestamp(t).isoformat()
                                       if ~np.isnan(t) else t for t in new_df['datetime_seconds']]
            new_df = new_df.drop(columns='datetime_seconds')

            new_df['cycle_index'] = cycle_index
            new_df['cycle_type'] = diag_cycle_type[diag_cycles_at.index(cycle_index)]
            new_df['step_index'] = step_index
            new_df['step_index_counter'] = step_index_counter
            new_df['step_type'] = diag_dict[cycle_index].index(step_index)
            new_df.astype({'cycle_index': 'int32',
                           'cycle_type': 'category',
                           'step_index': 'uint8',
                           'step_index_counter': 'int16',
                           'step_type': 'uint8'
                           })
            new_df['discharge_dQdV'] = new_df.discharge_capacity.diff() / new_df.voltage.diff()
            new_df['charge_dQdV'] = new_df.charge_capacity.diff() / new_df.voltage.diff()
            all_dfs.append(new_df)

        # Ignore the index to avoid issues with overlapping voltages
        result = pd.concat(all_dfs, ignore_index=True)
        # Cycle_index gets a little weird about typing, so round it here
        result.cycle_index = result.cycle_index.round()

        result = result.astype(STRUCTURE_DTYPES['diagnostic_interpolated'])

        return result

    @classmethod
    def from_arbin_file(cls, path, validate=False):
        """
        Creates RawCyclerRun from an Arbin data file.

        Args:
            path (str): file path to data file
            validate (bool): True if data is to be validated.

        Returns:
            beep.structure.RawCyclerRun
        """
        metadata_path = path.replace(".csv", "_Metadata.csv")
        data = pd.read_csv(path)
        data.rename(str.lower, axis='columns', inplace=True)

        for column, dtype in ARBIN_CONFIG['data_types'].items():
            if column in data:
                if not data[column].isnull().values.any():
                    data[column] = data[column].astype(dtype)

        data.rename(ARBIN_CONFIG['data_columns'], axis='columns', inplace=True)
        metadata = pd.read_csv(metadata_path)
        metadata.rename(str.lower, axis='columns', inplace=True)
        metadata.rename(ARBIN_CONFIG['metadata_fields'], axis='columns', inplace=True)
        # Note the to_dict, which scrubs numpy typing
        metadata = {col: item[0] for col, item
                    in metadata.to_dict('list').items()}

        # standardizing time format
        data['date_time_iso'] = data['date_time'].apply(
            lambda x: datetime.utcfromtimestamp(x).replace(tzinfo=pytz.UTC).isoformat())
        return cls(data, metadata, None, validate, filename=path)

    @classmethod
    def from_indigo_file(cls, path, validate=False):
        """
        Creates RawCyclerRun from an Indigo data file.

        Args:
            path (str): file path to data file
            validate (bool): True if data is to be validated.

        Returns:
            beep.structure.RawCyclerRun
        """

        data = pd.read_hdf(path, 'time_series_data')
        metadata = dict()

        if len(list(data['cell_id'].unique())) > 1:
            raise ValueError('More than 1 cell_id exists in {}'.format(path))

        metadata['indigo_cell_id'] = int(data['cell_id'].iloc[0])
        metadata['filename'] = path
        metadata['_today_datetime'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        # transformations
        data = data.reset_index().reset_index()  # twice in case old index is stored in file
        data = data.drop(columns=['index'])
        data.rename(columns={'level_0': 'data_point'}, inplace=True)
        data.loc[data.half_cycle_count % 2 == 1, 'charge_capacity'] = abs(data.cell_coulomb_count_c) / 3600
        data.loc[data.half_cycle_count % 2 == 0, 'charge_capacity'] = 0
        data.loc[data.half_cycle_count % 2 == 0, 'discharge_capacity'] = abs(data.cell_coulomb_count_c) / 3600
        data.loc[data.half_cycle_count % 2 == 1, 'discharge_capacity'] = 0
        data.loc[data.half_cycle_count % 2 == 1, 'charge_energy'] = abs(data.cell_energy_j)
        data.loc[data.half_cycle_count % 2 == 0, 'charge_energy'] = 0
        data.loc[data.half_cycle_count % 2 == 0, 'discharge_energy'] = abs(data.cell_energy_j)
        data.loc[data.half_cycle_count % 2 == 1, 'discharge_energy'] = 0
        data['internal_resistance'] = data.cell_voltage_v / data.cell_current_a
        data['date_time_iso'] = data['system_time_us']\
            .apply(lambda x: datetime.utcfromtimestamp(x/1000000).replace(tzinfo=pytz.UTC).isoformat())

        data.rename(INDIGO_CONFIG['data_columns'], axis='columns', inplace=True)

        metadata['start_datetime'] = data.sort_values(by='system_time_us')['date_time_iso'].iloc[0]

        return cls(data, metadata, None, validate, filename=path)

    @classmethod
    def from_biologic_file(cls, path, validate=False):
        """
        Creates RawCyclerRun from an Biologic data file.

        Args:
            path (str): file path to data file
            validate (bool): True if data is to be validated.

        Returns:
            beep.structure.RawCyclerRun
        """

        header_line = 3             # specific to file layout
        data_starts_line = 4        # specific to file layout
        column_map = BIOLOGIC_CONFIG['data_columns']

        raw = dict()
        i = 0
        with open(path, 'rb') as f:

            empty_lines = 0         # used to find the end of the data entries.
            while empty_lines < 2:  # finding 2 empty lines in a row => EOF
                line = f.readline()
                i += 1

                if i == header_line:
                    header = str(line.strip())[2:-1]
                    columns = header.split('\\t')
                    for c in columns:
                        raw[c] = list()

                if i >= data_starts_line:
                    line = line.strip().decode()
                    if len(line) == 0:
                        empty_lines += 1
                        continue
                    items = line.split('\t')
                    for ci in range(len(items)):
                        column_name = columns[ci]
                        data_type = column_map.get(column_name, dict()).get('data_type', str)
                        scale = column_map.get(column_name, dict()).get('scale', 1.0)
                        item = items[ci]
                        if data_type == 'int':
                            item = int(float(item))
                        if data_type == 'float':
                            item = float(item) * scale
                        raw[column_name].append(item)

        data = dict()
        for column_name in column_map.keys():
            data[column_map[column_name]['beep_name']] = raw[column_name]
        data['data_point'] = list(range(1, len(raw['cycle number']) + 1))

        data = pd.DataFrame(data)
        data.loc[data.step_index % 2 == 0, 'charge_capacity'] = abs(data.charge_capacity)
        data.loc[data.step_index % 2 == 1, 'charge_capacity'] = 0
        data.loc[data.step_index % 2 == 1, 'discharge_capacity'] = abs(data.discharge_capacity)
        data.loc[data.step_index % 2 == 0, 'discharge_capacity'] = 0

        metadata = dict()
        metadata['filename'] = path
        metadata['_today_datetime'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        return cls(data, metadata, None, validate, filename=path)

    @staticmethod
    def get_maccor_quantity_sum(data, quantity, state_type):
        """
        Computes non-decreasing capacity or energy (either charge or discharge)
        through multiple steps of a single cycle and resets capacity at the
        start of each new cycle. Input Maccor data resets to zero at each step.

        Args:
            data (pd.DataFrame): maccor data.
            quantity (str):  capacity or energy.
            state_type (str): charge or discharge.

        Returns:
            Series: summed quantities.
        """
        state_code = MACCOR_CONFIG["{}_state_code".format(state_type)]
        quantity_agg = data['_' + quantity].where(data['_state'] == state_code, other=0.0, axis=0)
        end_step = data['_ending_status'].apply(lambda x: MACCOR_CONFIG['end_step_code_min'] <= x
                                                          <= MACCOR_CONFIG['end_step_code_max'])
        end_step_inds = end_step.index[end_step]

        if end_step_inds.size == 0:
            return quantity_agg

        lastindex = quantity_agg.size - 1
        for i, istep in enumerate(end_step_inds):
            if i > 0:
                quantity_agg[istep_old+1:istep+1] += cycle_sum
            if istep == lastindex:
                cycle_sum = 0.
            elif data.loc[istep+1, 'cycle_index'] != data.loc[istep, 'cycle_index']:
                cycle_sum = 0.
            else:
                cycle_sum = quantity_agg[istep]
            istep_old = istep
        if end_step_inds[-1] < lastindex:
            quantity_agg[istep_old+1:] += cycle_sum
        return quantity_agg

    @classmethod
    def from_maccor_file(cls, filename, include_eis=True, validate=False):
        """
        Method for ingestion of Maccor format files.

        Args:
            filename (str): file path for maccor format file.
            include_eis (bool): whether to include the eis spectrum
                in the ingestion procedure.
            validate (bool): whether to validate on instantiation.
        """
        with open(filename) as f:
            metadata_line = f.readline().strip()

        # Parse data
        data = pd.read_csv(filename, delimiter="\t", skiprows=1)
        data.rename(str.lower, axis='columns', inplace=True)
        data = data.astype(MACCOR_CONFIG['data_types'])
        data.rename(MACCOR_CONFIG['data_columns'], axis='columns', inplace=True)
        data['charge_capacity'] = cls.get_maccor_quantity_sum(data, 'capacity', 'charge')
        data['discharge_capacity'] = cls.get_maccor_quantity_sum(data, 'capacity', 'discharge')
        data['charge_energy'] = cls.get_maccor_quantity_sum(data, 'energy', 'charge')
        data['discharge_energy'] = cls.get_maccor_quantity_sum(data, 'energy', 'discharge')

        if 'temperature' not in data.columns:
            data['temperature'] = np.NaN

        # Parse metadata - kinda hackish way to do it, but it works
        metadata = parse_maccor_metadata(metadata_line)
        metadata = pd.DataFrame(metadata)
        _, channel_number = os.path.splitext(filename)
        metadata['channel_id'] = int(channel_number.replace('.', ''))
        metadata.rename(str.lower, axis='columns', inplace=True)
        metadata.rename(MACCOR_CONFIG['metadata_fields'], axis='columns', inplace=True)
        # Note the to_dict, which scrubs numpy typing
        metadata = {col: item[0] for col, item
                    in metadata.to_dict('list').items()}

        # Check for EIS files
        if include_eis:
            eis_pattern = ".*.".join(filename.rsplit('.', 1))
            all_eis_files = glob(eis_pattern)
            eis = EISpectrum.from_maccor_file(all_eis_files[0])
        else:
            eis = None

        # standardizing time format
        data['date_time_iso'] = data['date_time'].apply(maccor_timestamp)

        return cls(data, metadata, eis, validate, filename=filename)

    def determine_structuring_parameters(self, v_range=None, resolution=1000,
                                         nominal_capacity=1.1, full_fast_charge=0.8,
                                         parameters_path='data-share/raw/parameters'):
        """
        Method for determining what values to use to convert raw run into processed run

        Args:
            v_range ([float, float]): voltage range for interpolation
            resolution (int): resolution for interpolation
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            parameters_path (str): path to parameters file

        Returns:
            v_range ([float, float]): voltage range for interpolation
            resolution (int): resolution for interpolation
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            diagnostic_available (dict): dictionary of values to use for
                finding and using the diagnostic cycles

        """
        run_parameter, all_parameters = get_protocol_parameters(self.filename, parameters_path)
        # Logic for interpolation variables and diagnostic cycles
        diagnostic_available = False
        if run_parameter is not None:
            if {'capacity_nominal'}.issubset(run_parameter.columns.tolist()):
                nominal_capacity = run_parameter['capacity_nominal'].iloc[0]
            if {'discharge_cutoff_voltage', 'charge_cutoff_voltage'}.issubset(run_parameter.columns):
                v_range = [all_parameters['discharge_cutoff_voltage'].min(),
                           all_parameters['charge_cutoff_voltage'].max()]
            if {'diagnostic_type', 'diagnostic_start_cycle', 'diagnostic_interval'}.issubset(run_parameter.columns):
                if run_parameter['diagnostic_type'].iloc[0] == 'HPPC+RPT':
                    hppc_rpt = ['reset', 'hppc', 'rpt_0.2C', 'rpt_1C', 'rpt_2C']
                    hppc_rpt_len = 5
                    diagnostic_starts_at = [1, 1 + run_parameter['diagnostic_start_cycle'].iloc[0] + 1 * hppc_rpt_len]
                    for i in range(1, 100):
                        diag_cycle_num = (i * (run_parameter['diagnostic_interval'].iloc[0] + hppc_rpt_len) +
                                          1 + run_parameter['diagnostic_start_cycle'].iloc[0] + 1 * hppc_rpt_len)
                        diagnostic_starts_at.append(diag_cycle_num)
                    diagnostic_available = {"parameter_set": run_parameter['diagnostic_parameter_set'].iloc[0],
                                            "cycle_type": hppc_rpt,
                                            "length": hppc_rpt_len,
                                            "diagnostic_starts_at": diagnostic_starts_at}

        return v_range, resolution, nominal_capacity, full_fast_charge, diagnostic_available

    def to_processed_cycler_run(self):
        """
        Method for converting to ProcessedCyclerRun

        Returns:
            beep.structure.ProcessedCyclerRun: ProcessedCyclerRun
                that corresponds to processed RawCyclerRun

        """
        v_range, resolution, nominal_capacity, full_fast_charge, diagnostic_available = \
            self.determine_structuring_parameters()

        return ProcessedCyclerRun.from_raw_cycler_run(
            self, v_range=v_range, resolution=resolution,
            nominal_capacity=nominal_capacity,
            full_fast_charge=full_fast_charge,
            diagnostic_available=diagnostic_available
        )

    def save_numpy_binary(self, name):
        """
        Save RawCyclerRun as a numeric array and metadata json

        Args:
            name (str): file prefix, saves to a .npz and .json file
        """
        float_array = np.array(self.data[self.FLOAT_COLUMNS].astype(np.float64))
        int_array = np.array(self.data[self.INT_COLUMNS].astype(np.int64))
        np.savez_compressed(name, float_array=float_array, int_array=int_array)
        dumpfn(self.metadata, "{}.json".format(name))

    @classmethod
    def load_numpy_binary(cls, name):
        """
        Load RawCyclerRun from numeric binary file

        Args:
            name (str): str prefix for numeric and metadata files

        Returns:
            beep_structure.RawCyclerRun loaded from binary files

        """
        loaded = np.load("{}.npz".format(name))
        data = dict(zip(cls.FLOAT_COLUMNS, np.transpose(loaded['float_array'])))
        data.update(dict(zip(cls.INT_COLUMNS, np.transpose(loaded['int_array']))))
        data = pd.DataFrame(data)
        metadata = loadfn("{}.json".format(name))
        return cls(data, metadata)


class ProcessedCyclerRun(MSONable):
    """
    Processed cycler run file which is intended to reflect the old format,
    for facile use in featurization, fitting, etc.  Note that this
    purposefully departs from PEP 8 guidelines to maintain certain
    naming tropes from the legacy code, which may be updated in the future.

    Attributes:
        barcode (str): barcode for the experiment.
        protocol (str): protocol for the experiment.
        channel_id (int): id for the channel for the experiment.
        summary (pandas.DataFrame): data of summary data for each cycle.
        cycles_interpolated (pandas.DataFrame): interpolated data for
            discharge over 2.8-3.5.
    """
    def __init__(self, barcode, protocol, channel_id, summary,
                 cycles_interpolated, diagnostic_summary=None,
                 diagnostic_interpolated=None):
        """
        For the most part, this invocation method will not be
        used directly, since this object will be invoked either
        from a RawCyclerRun object or a file.

        Args:
            barcode (string): barcode for the experiment
            protocol (string): protocol for the experiment
            channel_id (int): id for the channel for the experiment
            summary (pandas.DataFrame): data of summary data for each cycle
            cycles_interpolated (pandas.DataFrame): interpolated data for
                regular cycles
        """
        self.barcode = barcode
        self.protocol = protocol
        self.channel_id = channel_id
        self.summary = summary

        # We can drop this restriction later if we don't need it
        min_index = cycles_interpolated.cycle_index.min()
        if 'step_type' in cycles_interpolated.columns:
            min_index_df = cycles_interpolated[(cycles_interpolated.cycle_index == min_index) &
                                               (cycles_interpolated.step_type == 'discharge')]
        else:
            min_index_df = cycles_interpolated[(cycles_interpolated.cycle_index == min_index)]
        self.v_interpolated = min_index_df.voltage.values

        self.cycles_interpolated = cycles_interpolated
        self.diagnostic_summary = diagnostic_summary
        self.diagnostic_interpolated = diagnostic_interpolated

    @classmethod
    def from_raw_cycler_run(cls, raw_cycler_run, v_range=None, resolution=1000,
                            diagnostic_resolution=500, nominal_capacity=1.1,
                            full_fast_charge=0.8, diagnostic_available=False):
        """
        Method to invoke ProcessedCyclerRun from RawCyclerRun object

        Args:
            raw_cycler_run (beep.structure.RawCyclerRun): RawCyclerRun
                object to create ProcessedCyclerRun from.
            v_range ([int, int]): range of voltages for cycle interpolation.
            resolution (int): resolution for cycle interpolation.
            diagnostic_resolution (int): number of datapoints per step for
                interpolating diagnostic cycles.
            nominal_capacity (float): nominal capacity for summary stats.
            full_fast_charge (float): full fast charge for summary stats.
            diagnostic_available (dict): project metadata for processing
                diagnostic cycles correctly.
        """
        if diagnostic_available:
            diagnostic_summary = raw_cycler_run.get_diagnostic_summary(
                diagnostic_available)
            diagnostic_interpolated = raw_cycler_run.get_interpolated_diagnostic_cycles(
                diagnostic_available, diagnostic_resolution)
        else:
            diagnostic_summary = None
            diagnostic_interpolated = None

        cycles_interpolated = raw_cycler_run.get_interpolated_cycles(
            v_range=v_range, resolution=resolution, diagnostic_available=diagnostic_available)
        return cls(raw_cycler_run.metadata.get("barcode"),
                   raw_cycler_run.metadata.get("protocol"),
                   raw_cycler_run.metadata.get("channel_id"),
                   raw_cycler_run.get_summary(
                       nominal_capacity=nominal_capacity,
                       full_fast_charge=full_fast_charge
                   ),
                   cycles_interpolated,
                   diagnostic_summary,
                   diagnostic_interpolated)

    @classmethod
    def auto_load(cls, filename, validate=False):
        """
        Method for loading processed cycler run from raw cycler filename,
        processing it according to prescribed logic corresponding to the
        file name and or the file's contents.

        Args:
            filename (str): filename associated with the project
            validate (bool): whether or not to validate file

        Returns:
            beep.structure.ProcessedCyclerRun: ProcessedCyclerRun corresponding
                to the read and processed data from the filename

        """
        # Arbin files are via standard pipeline
        if re.match(FastCharge_CONFIG['file_pattern'], filename):
            raw = RawCyclerRun.from_arbin_file(filename, validate)
            return raw.to_processed_cycler_run()
        elif re.match(xTesladiag_CONFIG['file_pattern'], filename):
            raw = RawCyclerRun.from_maccor_file(filename, validate)
            return raw.to_processed_cycler_run()

        else:
            raise ValueError("File pattern or contents of {} not recognized".format(
                filename
            ))

    def get_cycle_life(self, n_cycles_cutoff=40, threshold=0.8):
        """
        Calculate cycle life for capacity loss below a certain threshold

        Args:
            n_cycles_cutoff (int): cutoff for number of cycles to sample
                for the cycle life in order to use median method.
            threshold (float): fraction of capacity loss for which
                to find the cycle index.

        Returns:
            float: cycle life.
        """
        # discharge_capacity has a spike and  then increases slightly between \
        # 1-40 cycles, so let us use take median of 1st 40 cycles for max.

        # If n_cycles <  n_cycles_cutoff, do not use median method
        if len(self.summary) > n_cycles_cutoff:
            max_capacity = np.median(self.summary.discharge_capacity.iloc[0:n_cycles_cutoff])
        else:
            max_capacity = 1.1

        # If capacity falls below 80% of initial capacity by end of run
        if self.summary.discharge_capacity.iloc[-1] / max_capacity <= threshold:
            cycle_life = self.summary[self.summary.discharge_capacity < threshold * max_capacity].index[0]
        else:
            # Some cells do not degrade below the threshold (low degradation rate)
            cycle_life = len(self.summary) + 1

        return cycle_life

    def capacities_at_set_cycles(self, cycle_min=200, cycle_max=1800, cycle_interval=200):
        """
        Get discharge capacity at constant intervals of 200 cycles

        Args:
            cycle_min (int): Cycle number to being forecasting capacity at
            cycle_max (int): Cycle number to end forecasting capacity at
            cycle_interval (int): Intervals for forecasts

        Returns:
            pandas.DataFrame:
        """
        discharge_capacities = pd.DataFrame(np.zeros((1, int((cycle_max-cycle_min)/cycle_interval))))
        counter = 0
        cycle_indices = np.arange(cycle_min, cycle_max, cycle_interval)
        for cycle_index in cycle_indices:
            try:
                discharge_capacities[counter] = self.summary.discharge_capacity.iloc[cycle_index]
            except IndexError:
                pass
            counter = counter + 1
        discharge_capacities.columns = np.core.defchararray.add("cycle_", cycle_indices.astype(str))
        return discharge_capacities

    def cycles_to_reach_set_capacities(self, thresh_max_cap=0.98, thresh_min_cap=0.78, interval_cap=0.03):
        """
        Get cycles to reach set threshold capacities.

        Args:
            thresh_max_cap (float): Upper bound on capacity to compute cycles at.
            thresh_min_cap (float): Lower bound on capacity to compute cycles at.
            interval_cap (float): Interval/step size.

        Returns:
            pandas.DataFrame:
        """
        threshold_list = np.around(np.arange(thresh_max_cap, thresh_min_cap, - interval_cap), 2)
        counter = 0
        cycles = pd.DataFrame(np.zeros((1, len(threshold_list))))
        for threshold in threshold_list:
            cycles[counter] = self.get_cycle_life(threshold=threshold)
            counter = counter + 1
        cycles.columns = np.core.defchararray.add("capacity_", threshold_list.astype(str))
        return cycles

    def as_dict(self):
        """
        Method for dictionary serialization.

        Returns:
            dict: corresponding to dictionary for serialization.

        """
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "barcode": self.barcode,
                "protocol": self.protocol,
                "channel_id": self.channel_id,
                "summary": self.summary.to_dict("list"),
                "cycles_interpolated": self.cycles_interpolated.to_dict("list"),
                "diagnostic_summary":
                    self.diagnostic_summary.to_dict("list") if self.diagnostic_summary is not None else None,
                "diagnostic_interpolated":
                    self.diagnostic_interpolated.to_dict("list") if self.diagnostic_interpolated is not None else None
                }

    @classmethod
    def from_dict(cls, d):
        """

        Args:
            d (dict): dictionary represenation.

        Returns:
            beep.structure.ProcessedCyclerRun: deserialized ProcessedCyclerRun.
        """
        """MSONable deserialization method"""
        for obj, dtype_dict in STRUCTURE_DTYPES.items():
            for column, dtype in dtype_dict.items():
                if d.get(obj) is not None:
                    if d[obj].get(column) is not None:
                        d[obj][column] = pd.Series(d[obj][column], dtype=dtype)

        d['cycles_interpolated'] = pd.DataFrame(d['cycles_interpolated'])
        d['summary'] = pd.DataFrame(d['summary'])
        d['diagnostic_summary'] = pd.DataFrame(d.get('diagnostic_summary'))
        d['diagnostic_interpolated'] = pd.DataFrame(d.get('diagnostic_interpolated'))
        return cls(**d)

    METADATA_ATTRIBUTE_ORDER = ['barcode', 'protocol', 'channel_id']
    SUMMARY_COLUMN_ORDER = ['discharge_capacity', 'charge_capacity', 'dc_internal_resistance', 'temperature_maximum',
                            'temperature_average', 'temperature_minimum', "charge_duration"]
    CYCLES_INTERPOLATED_COLUMN_ORDER = ['cycle_index', 'voltage', 'current', 'internal_resistance',
                                        'charge_capacity', 'discharge_capacity', 'temperature']

    def save_numpy_binary(self, name):
        """
        Save ProcessedCyclerRun as numpy binary.

        Args:
            name (str): filename to save numpy binary as.
        """
        meta_array = np.array([getattr(self, mattribute) for mattribute
                               in self.METADATA_ATTRIBUTE_ORDER])
        summary_array = self.summary[self.SUMMARY_COLUMN_ORDER].values
        cycles_interpolated_array = self.cycles_interpolated[
            self.CYCLES_INTERPOLATED_COLUMN_ORDER].values
        np.savez_compressed(name, meta=meta_array, summary=summary_array,
                            cycles_interpolated=cycles_interpolated_array)

    @classmethod
    def load_numpy_binary(cls, name):
        """
        Class method to load ProcessedCyclerRun from numpy binary.

        Args:
            name (str): filename for numpy binary to be loaded.

        Returns:
            beep.structure.ProcessedCyclerRun: loaded from numpy binary

        """
        if not name.endswith(".npz"):
            name += ".npz"
        data = np.load(name, allow_pickle=True)
        meta_kwargs = dict(zip(cls.METADATA_ATTRIBUTE_ORDER, data['meta']))

        # Load summary into DataFrame
        summary = pd.DataFrame(dict(zip(cls.SUMMARY_COLUMN_ORDER,
                                        data['summary'].transpose())))

        # Load cycles_interpolated into DataFrame
        cycles_interpolated = pd.DataFrame(
            dict(zip(cls.CYCLES_INTERPOLATED_COLUMN_ORDER,
                     data['cycles_interpolated'].transpose())))
        return cls(summary=summary, cycles_interpolated=cycles_interpolated,
                   **meta_kwargs)


class EISpectrum(MSONable):
    """
    Class describing an Electrochemical Impedance Spectrum
    """
    def __init__(self, data, metadata):
        """

        Args:
            data:
            metadata:
        """
        self.data = data
        self.metadata = metadata

    @classmethod
    def from_csv(cls, filename):
        """

        Args:
            filename(str): path to data file.

        Returns:
            beep.structure.EISpectrum: EISpectrum object representation of
                data.
        """
        raise NotImplementedError("from_csv not implemented for EISpectrum")

    @classmethod
    def from_maccor_file(cls, filename):
        """
        EISpectrum from Maccor file.

        Args:
            filename (str): file path to data.

        Returns:
            beep.strucure.EISpectrum: EISpectrum object representation of
                data.
        """
        with open(filename) as f:
            lines = f.readlines()
        # Parse freq sweep, method, and output filename
        freq_sweep = lines[1].split('Frequency Sweep:')[1].strip()
        freq_sweep = freq_sweep.replace('Circut', "Circuit")
        method = lines[2]
        filename = lines[3]

        # Parse start datetime and reformat in isoformat
        start = lines[6].split('Start Date:')[1].strip()
        date, time = start.split('Start Time:')
        start = ','.join([date.strip(), time.strip()])
        start = datetime.strptime(start, "%A, %B %d, %Y,%H:%M")
        start = start.isoformat()

        line_8 = lines[8].split()

        # Construct metadata dictionary
        metadata = {"frequency_sweep": freq_sweep,
                    "method": method,
                    "filename": filename,
                    "start": start,
                    "line_8": line_8}

        data = '\n'.join(lines[10:])
        data = pd.read_csv(StringIO(data), delimiter="\t")

        return cls(data=data, metadata=metadata)

    def as_dict(self):
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "data": self.data.to_dict("list"),
                "metadata": self.metadata.to_dict()}

    @classmethod
    def from_dict(cls, d):
        data = pd.DataFrame(d['data'])
        data = data.sort_index()
        metadata = pd.DataFrame(d['metadata'])
        return cls(data, metadata)


def determine_whether_step_is_discharging(step_dataframe):
    """
    Helper function to determine whether a given dataframe corresponding
    to a single cycle_index/step is charging or discharging, only intended
    to be used with a dataframe for single step/cycle

    Args:
         step_dataframe (pandas.DataFrame): dataframe to determine whether
         charging or discharging
    """
    cap = step_dataframe[["charge_capacity", "discharge_capacity"]]
    return cap.diff(axis=0).mean(axis=0).diff().iloc[-1] > 0


def determine_whether_step_is_charging(step_dataframe):
    """
    Helper function to determine whether a given dataframe corresponding
    to a single cycle_index/step is charging or discharging, only intended
    to be used with a dataframe for single step/cycle

    Args:
         step_dataframe (pandas.DataFrame): dataframe to determine whether
         charging or discharging
    """
    cap = step_dataframe[["charge_capacity", "discharge_capacity"]]
    return cap.diff(axis=0).mean(axis=0).diff().iloc[-1] < 0


def get_interpolated_data(dataframe, field_name='voltage', field_range=None,
                          columns=None, resolution=1000):
    """
    General method for creating uniform (i. e. same # of data points)
    dataframe using an by interpolation of supplied columns on the
    specified field, assumes input field data is monotonic

    Args:
        dataframe (pandas.DataFrame): dataframe to create interpolate
        field_name (str): column name to use as the dependent interpolation variable
        field_range (list): list of two values to use as endpoints, if None,
            range is the min/max of the dataframe field_name values
        columns (list): list of column names to provide interpolated values for,
            default value of None indicates all columns should be interpolated
        resolution (int): number of data points to sample in the uniform
            cycles, defaults to 1000

    Returns:
        pandas.DataFrame: DataFrame of interpolated values
    """
    columns = columns or dataframe.columns
    columns = list(set(columns) | {field_name})

    df = dataframe.loc[:, columns]
    field_range = field_range or [df[field_name].iloc[0], df[field_name].iloc[-1]]
    # If interpolating on datetime, change column to datetime series and
    # use date_range to create interpolating vector
    if field_name == 'date_time_iso':
        df['date_time_iso'] = pd.to_datetime(df['date_time_iso'])
        interp_x = pd.date_range(
            start=df[field_name].iloc[0], end=df[field_name].iloc[-1], periods=resolution)
    else:
        interp_x = np.linspace(*field_range, resolution)
    interpolated_df = pd.DataFrame({field_name: interp_x, "interpolated": True})

    df['interpolated'] = False

    # Merge interpolated and uninterpolated DFs to use pandas interpolation
    interpolated_df = interpolated_df.merge(df, how='outer', on=field_name, sort=True)
    interpolated_df = interpolated_df.set_index(field_name)
    interpolated_df = interpolated_df.interpolate('slinear')

    # Filter for only interpolated values
    interpolated_df[['interpolated_x']] = interpolated_df[['interpolated_x']].fillna(False)
    interpolated_df = interpolated_df[interpolated_df['interpolated_x']]
    interpolated_df = interpolated_df.drop(["interpolated_x", "interpolated_y"], axis=1)
    interpolated_df = interpolated_df.reset_index()
    # Remove duplicates
    interpolated_df = interpolated_df[~interpolated_df[field_name].duplicated()]
    return interpolated_df


def diagnostic_function(df, column):
    """

    Args:
        df (pandas.DataFrame): A dataframe.
        column (str): A column name.

    Returns:
        float: median value of column.
    """
    value = df[column].agg('median')
    return value


def parse_maccor_metadata(metadata_string):
    """
    Parses maccor metadata string, which is annoyingly inconsistent.
    Basically just splits the string by a set of fields and creates
    a dictionary of pairs of fields and values with colons scrubbed
    from fields.

    Args:
        metadata_string (str): string corresponding to maccor metadata.

    Returns:
        dict: dictionary of metadata fields and values.

    """
    metadata_fields = ['Today\'s Date',
                       'Date of Test:',
                       'Filename:',
                       'Procedure:',
                       'Comment/Barcode:']
    metadata_values = split_string_by_fields(metadata_string, metadata_fields)
    metadata = {k.replace(':', ''): [v.strip()]
                for k, v in zip(metadata_fields, metadata_values)}
    return metadata


def get_project_sequence(path):
    """
    Returns project sequence for a given path

    Args:
        path (str): full project file path

    Returns:
        ([str]): list of project parts

    """
    root, file = os.path.split(path)
    file_parts = file.split('_')
    return file_parts


def get_protocol_parameters(filepath, parameters_path='data-share/raw/parameters'):
    """
    Helper function to get the project parameters for a file given the filename

    Args:
        filepath (str): full path to the file
        parameters_path (str): location to look for parameter files

    Returns:
        pandas.DataFrame: single row DataFrame corresponding to the parameters for this file
        pandas.DataFrame: DataFrame with all of the parameter for the project

    """
    project_name_list = get_project_sequence(filepath)
    project_name = project_name_list[0]
    path = os.path.join(os.environ.get("BEEP_PROCESSING_DIR", "/"), parameters_path)
    project_parameter_files = glob(os.path.join(path, project_name + '*'))
    assert len(project_parameter_files) <= 1, 'Found too many parameter files for: ' + project_name

    if len(project_parameter_files) == 1:
        df = pd.read_csv(project_parameter_files[0])
        parameter_row = df[df.seq_num == int(project_name_list[1])]
        if parameter_row.empty:
            logger.error("Unable to get project parameters for: %s", filepath, extra=s)
            parameter_row = None
            df = None
    else:
        parameter_row = None
        df = None
    return parameter_row, df


def get_diagnostic_parameters(diagnostic_available, diagnostic_parameter_path,
                              project_name):
    """
    Interpolates data according to location and type of diagnostic
    cycles in the data

    Args:
        diagnostic_available (dict): dictionary with diagnostic_types as list,
            'length' of the diagnostic in cycles and location of the diagnostic
        diagnostic_parameter_path (str): full path to the location of the
            diagnostic parameter files
        project_name (str): name of the project to search with

    Returns:
        (list): containing upper and lower voltage limits for the
            diagnostic cycle

    """
    project_diag_files = glob(os.path.join(diagnostic_parameter_path, project_name + '*'))
    assert len(project_diag_files) <= 1, 'Found too many diagnostic parameter files for: ' + \
                                         project_name

    # Find the voltage range for the diagnostic cycles
    if len(project_diag_files) == 1:
        df = pd.read_csv(project_diag_files[0])
        diag_row = df[df.diagnostic_parameter_set == diagnostic_available['parameter_set']]
        v_range = [diag_row['diagnostic_discharge_cutoff_voltage'].iloc[0],
                   diag_row['diagnostic_charge_cutoff_voltage'].iloc[0]]
    else:
        v_range = [2.7, 4.2]

    return v_range


def split_string_by_fields(string, fields):
    """
    Helper function to split a string by a set of ordered strings,
    primarily used for Maccor metadata parsing.

    >>>split_string_by_fields("first name: Joey  last name Montoya",
    >>>                       ["first name:", "last name"])
    ["Joey", "Montoya"]

    Args:
        string (str): string input to be split
        fields (list): list of fields to split input string by.

    Returns:
        list: substrings corresponding to the split input strings.

    """
    # A bit brittle, there's probably something more clever with recursion
    substrings = []
    init, leftovers = string.split(fields[0])
    for field in fields[1:]:
        init, leftovers = leftovers.split(field)
        substrings.append(init)
    substrings.append(leftovers)
    return substrings


def add_file_prefix_to_path(path, prefix):
    """
    Helper function to add file prefix to path.

    Args:
        path (str): full path to file.
        prefix (str): prefix for file.

    Returns:
        str: path with prefix appended to filename.

    """
    split_path = list(os.path.split(path))
    split_path[-1] = prefix + split_path[-1]
    return os.path.join(*split_path)


def determine_paused(group, paused_threshold=3600):
    """
    Evaluate a raw cycling dataframe to determine if there is a pause in cycling

    Args:
        group (pd.DataFrame): cycling dataframe with date_time_iso column
        paused_threshold (int): gap in seconds to classify as a pause in cycling

    Returns:
        bool: is there a pause in this cycle?

    """
    date_time_obj = pd.to_datetime(group['date_time_iso'])
    date_time_float = [time.mktime(t.timetuple())
                       if t is not pd.NaT else float('nan')
                       for t in date_time_obj]
    date_time_float = pd.Series(date_time_float)
    return int(date_time_float.diff().max() > paused_threshold)


def maccor_timestamp(x):
    """
    Helper function with exception handling for cases where the
    maccor cycler mis-prints the datetime stamp for the row. This
    happens when data is being recorded rapidly as the date switches over
    ie. between 10/21/2019 23:59:59 and 10/22/2019 00:00:00.

    Args:
        x (str): The datetime string for maccor in format '%m/%d/%Y %H:%M:%S'

    Returns:
        datetime.Datetime: Datetime object in iso format (daylight savings aware)

    """
    pacific = pytz.timezone('US/Pacific')
    utc = pytz.timezone('UTC')
    try:
        iso = pacific.localize(datetime.strptime(x, '%m/%d/%Y %H:%M:%S'),
                               is_dst=True).astimezone(utc).isoformat()
    except ValueError:
        x = x + ' 00:00:00'
        iso = pacific.localize(datetime.strptime(x, '%m/%d/%Y %H:%M:%S'),
                               is_dst=True).astimezone(utc).isoformat()
    return iso


def process_file_list_from_json(file_list_json, processed_dir='data-share/structure/'):
    """
    Function to take a json filename corresponding to a data structure
    with a 'file_list' and a 'validity' attribute, process each file
    with a corresponding True validity, dump the processed file into
    a predetermined directory, and return a jsonable dict of processed
    cycler run file locations

    Args:
        file_list_json (str): json string or json filename corresponding
            to a dictionary with a file_list and validity attribute,
            if this string ends with ".json", a json file is assumed
            and loaded, otherwise interpreted as a json string.
        processed_dir (str): location for processed cycler run output
            files to be placed.

    Returns:
        str: json string of processed files (with key "processed_file_list").
            Note that this list contains None values for every file that
            had a corresponding False in the validity list.

    """
    # Get file list and validity from json, if ends with .json,
    # assume it's a file, if not assume it's a json string
    if file_list_json.endswith(".json"):
        file_list_data = loadfn(file_list_json)
    else:
        file_list_data = json.loads(file_list_json)

    # Setup Events
    events = KinesisEvents(service='DataStructurer', mode=file_list_data['mode'])

    # Prepend optional root to output directory
    processed_dir = os.path.join(os.environ.get("BEEP_PROCESSING_DIR", "/"), processed_dir)

    if not os.path.exists(processed_dir):
        os.makedirs(processed_dir)

    file_list = file_list_data['file_list']
    validities = file_list_data['validity']
    run_ids = file_list_data['run_list']
    processed_file_list = []
    processed_run_list = []
    processed_result_list = []
    processed_message_list = []
    invalid_file_list = []
    for filename, validity, run_id in zip(file_list, validities, run_ids):
        logger.info('run_id=%s structuring=%s', str(run_id), filename, extra=s)
        if validity == 'valid':
            # Process raw cycler run and dump to file
            raw_cycler_run = RawCyclerRun.from_file(filename)
            processed_cycler_run = raw_cycler_run.to_processed_cycler_run()
            new_filename, ext = os.path.splitext(os.path.basename(filename))
            new_filename = new_filename + ".json"
            new_filename = add_suffix_to_filename(new_filename, "_structure")
            processed_cycler_run_loc = os.path.join(processed_dir, new_filename)
            processed_cycler_run_loc = os.path.abspath(processed_cycler_run_loc)
            dumpfn(processed_cycler_run, processed_cycler_run_loc)

            # Append file loc to list to be returned
            processed_file_list.append(processed_cycler_run_loc)
            processed_run_list.append(run_id)
            processed_result_list.append("success")
            processed_message_list.append({'comment': '',
                                           'error': ''})

        else:
            invalid_file_list.append(filename)

    output_json = {"file_list": processed_file_list,
                   "run_list": processed_run_list,
                   "result_list": processed_result_list,
                   "message_list": processed_message_list,
                   "invalid_file_list": invalid_file_list}

    events.put_structuring_event(output_json, 'complete')

    # Return jsonable file list
    return json.dumps(output_json)


def main():
    """
    Main function of this module, takes in arguments of an input
    and output filename and uses the input file to create a
    structured data output for analysis/ML processing.
    """
    logger.info('starting', extra=s)
    logger.info('Running version=%s', __version__, extra=s)
    try:
        args = docopt(__doc__)
        input_json = args['INPUT_JSON']
        print(process_file_list_from_json(input_json))
    except Exception as e:
        logger.error(str(e), extra=s)
        raise e
    logger.info('finish', extra=s)
    return None


if __name__ == "__main__":
    main()
